# --- BEGIN COPYRIGHT BLOCK ---
# Copyright (C) 2016-2017 Red Hat, Inc.
# All rights reserved.
#
# License: GPL (version 3 or any later version).
# See LICENSE for details.
# --- END COPYRIGHT BLOCK ---

import pytest

from lib389.tests.cli import topology as default_topology
from lib389.cli_base import LogCapture, FakeArgs
from lib389.plugins import RootDNAccessControlPlugin
from lib389.cli_conf.plugins import rootdn_ac as rootdn_cli


@pytest.fixture(scope="module")
def topology(request):
    topology = default_topology(request)

    plugin = RootDNAccessControlPlugin(topology.standalone)
    if not plugin.exists():
        plugin.create()

    # we need to restart the server after enabling the plugin
    plugin.enable()
    topology.standalone.restart()
    topology.logcap.flush()

    return topology

def test_set_open_time(topology):
    args = FakeArgs()

    args.value = "1030"
    rootdn_cli.set_open_time(topology.standalone, None, topology.logcap.log, args)
    assert topology.logcap.contains("rootdn-open-time set to")
    topology.logcap.flush()

    args.value = None
    rootdn_cli.display_time(topology.standalone, None, topology.logcap.log, args)
    assert topology.logcap.contains("rootdn-open-time: 1030")
    topology.logcap.flush()

def test_set_close_time(topology):
    args = FakeArgs()

    args.value = "1545"
    rootdn_cli.set_close_time(topology.standalone, None, topology.logcap.log, args)
    assert topology.logcap.contains("rootdn-close-time set to")
    topology.logcap.flush()

    args.value = None
    rootdn_cli.display_time(topology.standalone, None, topology.logcap.log, args)
    assert topology.logcap.contains("rootdn-close-time: 1545")
    topology.logcap.flush()

def test_clear_time(topology):
    args = FakeArgs()

    rootdn_cli.clear_time(topology.standalone, None, topology.logcap.log, args)
    assert topology.logcap.contains("time-based policy was cleared")
    topology.logcap.flush()

    args.value = None
    rootdn_cli.display_time(topology.standalone, None, topology.logcap.log, args)
    assert topology.logcap.contains("rootdn-open-time is not set")
    assert topology.logcap.contains("rootdn-close-time is not set")
    topology.logcap.flush()

def test_allow_ip(topology):
    args = FakeArgs()

    args.value = "127.0.0.1"
    rootdn_cli.allow_ip(topology.standalone, None, topology.logcap.log, args)
    assert topology.logcap.contains("{} added to rootdn-allow-ip".format(args.value))
    topology.logcap.flush()

    args.value = None
    rootdn_cli.display_ips(topology.standalone, None, topology.logcap.log, args)
    assert topology.logcap.contains("rootdn-allow-ip: 127.0.0.1")
    topology.logcap.flush()

def test_deny_ip(topology):
    args = FakeArgs()

    args.value = "127.0.0.2"
    rootdn_cli.deny_ip(topology.standalone, None, topology.logcap.log, args)
    assert topology.logcap.contains("{} added to rootdn-deny-ip".format(args.value))
    topology.logcap.flush()

    args.value = None
    rootdn_cli.display_ips(topology.standalone, None, topology.logcap.log, args)
    assert topology.logcap.contains("rootdn-deny-ip: 127.0.0.2")
    topology.logcap.flush()

def test_when_ip_is_allowed_its_not_denied(topology):
    args = FakeArgs()

    args.value = "127.0.0.3"
    rootdn_cli.deny_ip(topology.standalone, None, topology.logcap.log, args)
    assert topology.logcap.contains("{} added to rootdn-deny-ip".format(args.value))
    topology.logcap.flush()

    args.value = None
    rootdn_cli.display_ips(topology.standalone, None, topology.logcap.log, args)
    assert topology.logcap.contains("rootdn-deny-ip: 127.0.0.3")
    topology.logcap.flush()

    args.value = "127.0.0.3"
    rootdn_cli.allow_ip(topology.standalone, None, topology.logcap.log, args)
    assert topology.logcap.contains("{} added to rootdn-allow-ip".format(args.value))
    topology.logcap.flush()

    args.value = None
    rootdn_cli.display_ips(topology.standalone, None, topology.logcap.log, args)
    assert topology.logcap.contains("rootdn-allow-ip: 127.0.0.3")
    assert not topology.logcap.contains("rootdn-deny-ip: 127.0.0.3")
    topology.logcap.flush()

def test_when_ip_is_denied_its_not_allowed(topology):
    args = FakeArgs()

    args.value = "127.0.0.4"
    rootdn_cli.allow_ip(topology.standalone, None, topology.logcap.log, args)
    assert topology.logcap.contains("{} added to rootdn-allow-ip".format(args.value))
    topology.logcap.flush()

    args.value = None
    rootdn_cli.display_ips(topology.standalone, None, topology.logcap.log, args)
    assert topology.logcap.contains("rootdn-allow-ip: 127.0.0.4")
    topology.logcap.flush()

    args.value = "127.0.0.4"
    rootdn_cli.deny_ip(topology.standalone, None, topology.logcap.log, args)
    assert topology.logcap.contains("{} added to rootdn-deny-ip".format(args.value))
    topology.logcap.flush()

    args.value = None
    rootdn_cli.display_ips(topology.standalone, None, topology.logcap.log, args)
    assert topology.logcap.contains("rootdn-deny-ip: 127.0.0.4")
    assert not topology.logcap.contains("rootdn-allow-ip: 127.0.0.4")
    topology.logcap.flush()

def test_clear_ips(topology):
    args = FakeArgs()

    rootdn_cli.clear_all_ips(topology.standalone, None, topology.logcap.log, args)
    assert topology.logcap.contains("ip-based policy was cleared")
    topology.logcap.flush()

def test_allow_host(topology):
    args = FakeArgs()

    args.value = "example1.com"
    rootdn_cli.allow_host(topology.standalone, None, topology.logcap.log, args)
    assert topology.logcap.contains("{} added to rootdn-allow-host".format(args.value))
    topology.logcap.flush()

    args.value = None
    rootdn_cli.display_hosts(topology.standalone, None, topology.logcap.log, args)
    assert topology.logcap.contains("rootdn-allow-host: example1.com")
    topology.logcap.flush()

def test_deny_host(topology):
    args = FakeArgs()

    args.value = "example2.com"
    rootdn_cli.deny_host(topology.standalone, None, topology.logcap.log, args)
    assert topology.logcap.contains("{} added to rootdn-deny-host".format(args.value))
    topology.logcap.flush()

    args.value = None
    rootdn_cli.display_hosts(topology.standalone, None, topology.logcap.log, args)
    assert topology.logcap.contains("rootdn-deny-host: example2.com")
    topology.logcap.flush()

def test_when_host_is_allowed_its_not_denied(topology):
    args = FakeArgs()

    args.value = "example3.com"
    rootdn_cli.deny_host(topology.standalone, None, topology.logcap.log, args)
    assert topology.logcap.contains("{} added to rootdn-deny-host".format(args.value))
    topology.logcap.flush()

    args.value = None
    rootdn_cli.display_hosts(topology.standalone, None, topology.logcap.log, args)
    assert topology.logcap.contains("rootdn-deny-host: example3.com")
    topology.logcap.flush()

    args.value = "example3.com"
    rootdn_cli.allow_host(topology.standalone, None, topology.logcap.log, args)
    assert topology.logcap.contains("{} added to rootdn-allow-host".format(args.value))
    topology.logcap.flush()

    args.value = None
    rootdn_cli.display_hosts(topology.standalone, None, topology.logcap.log, args)
    assert topology.logcap.contains("rootdn-allow-host: example3.com")
    assert not topology.logcap.contains("rootdn-deny-host: example3.com")
    topology.logcap.flush()

def test_when_host_is_denied_its_not_allowed(topology):
    args = FakeArgs()

    args.value = "example4.com"
    rootdn_cli.allow_host(topology.standalone, None, topology.logcap.log, args)
    assert topology.logcap.contains("{} added to rootdn-allow-host".format(args.value))
    topology.logcap.flush()

    args.value = None
    rootdn_cli.display_hosts(topology.standalone, None, topology.logcap.log, args)
    assert topology.logcap.contains("rootdn-allow-host: example4.com")
    topology.logcap.flush()

    args.value = "example4.com"
    rootdn_cli.deny_host(topology.standalone, None, topology.logcap.log, args)
    assert topology.logcap.contains("{} added to rootdn-deny-host".format(args.value))
    topology.logcap.flush()

    args.value = None
    rootdn_cli.display_hosts(topology.standalone, None, topology.logcap.log, args)
    assert topology.logcap.contains("rootdn-deny-host: example4.com")
    assert not topology.logcap.contains("rootdn-allow-host: example4.com")
    topology.logcap.flush()

def test_clear_hosts(topology):
    args = FakeArgs()

    rootdn_cli.clear_all_hosts(topology.standalone, None, topology.logcap.log, args)
    assert topology.logcap.contains("host-based policy was cleared")
    topology.logcap.flush()

def test_allow_and_deny_days(topology):
    args = FakeArgs()

    args.value = "Mon".capitalize()
    rootdn_cli.allow_day(topology.standalone, None, topology.logcap.log, args)
    assert topology.logcap.contains("{} added to rootdn-days-allowed".format(args.value))
    topology.logcap.flush()

    args.value = None
    rootdn_cli.display_days(topology.standalone, None, topology.logcap.log, args)
    assert topology.logcap.contains("rootdn-days-allowed: Mon")
    topology.logcap.flush()

    args.value = "friday".capitalize()
    rootdn_cli.allow_day(topology.standalone, None, topology.logcap.log, args)
    assert topology.logcap.contains("Fri added to rootdn-days-allowed")
    topology.logcap.flush()

    args.value = None
    rootdn_cli.display_days(topology.standalone, None, topology.logcap.log, args)
    assert topology.logcap.contains("rootdn-days-allowed: Mon, Fri")
    topology.logcap.flush()

    args.value = "MONDAY".capitalize()
    rootdn_cli.deny_day(topology.standalone, None, topology.logcap.log, args)
    assert topology.logcap.contains("Mon removed from rootdn-days-allowed")
    topology.logcap.flush()

    args.value = None
    rootdn_cli.display_days(topology.standalone, None, topology.logcap.log, args)
    assert topology.logcap.contains("rootdn-days-allowed: Fri")
    topology.logcap.flush()

    args.value = "fri".capitalize()
    rootdn_cli.deny_day(topology.standalone, None, topology.logcap.log, args)
    assert topology.logcap.contains("Fri removed from rootdn-days-allowed")
    topology.logcap.flush()

    args.value = None
    rootdn_cli.display_days(topology.standalone, None, topology.logcap.log, args)
    assert topology.logcap.contains("No day-based access control policy has been configured")
    topology.logcap.flush()

def test_clear_days(topology):
    args = FakeArgs()

    rootdn_cli.clear_all_days(topology.standalone, None, topology.logcap.log, args)
    assert topology.logcap.contains("day-based policy was cleared")
    topology.logcap.flush()



# --- BEGIN COPYRIGHT BLOCK ---
# Copyright (C) 2016, William Brown <william at blackhats.net.au>
# All rights reserved.
#
# License: GPL (version 3 or any later version).
# See LICENSE for details.
# --- END COPYRIGHT BLOCK ---

from lib389.idm.organizationalunit import OrganizationalUnit, OrganizationalUnits, MUST_ATTRIBUTES
from lib389.cli_base import populate_attr_arguments, _generic_modify, CustomHelpFormatter
from lib389.cli_idm import (
    _generic_list,
    _generic_get,
    _generic_get_dn,
    _generic_create,
    _generic_rename,
    _generic_delete,
    _get_arg,
    _get_attributes,
    _warn,
    )

SINGULAR = OrganizationalUnit
MANY = OrganizationalUnits
RDN = 'ou'

# These are a generic specification, try not to tamper with them

def list(inst, basedn, log, args):
    _generic_list(inst, basedn, log.getChild('_generic_list'), MANY, args)

def get(inst, basedn, log, args):
    rdn = _get_arg( args.selector, msg="Enter %s to retrieve" % RDN)
    _generic_get(inst, basedn, log.getChild('_generic_get'), MANY, rdn, args)

def get_dn(inst, basedn, log, args):
    dn = _get_arg( args.dn, msg="Enter dn to retrieve")
    _generic_get_dn(inst, basedn, log.getChild('_generic_get_dn'), MANY, dn, args)

def create(inst, basedn, log, args):
    kwargs = _get_attributes(args, MUST_ATTRIBUTES)
    _generic_create(inst, basedn, log.getChild('_generic_create'), MANY, kwargs, args)

def delete(inst, basedn, log, args, warn=True):
    dn = _get_arg(args.dn, msg="Enter dn to delete")
    if warn:
        _warn(dn, msg="Deleting %s %s" % (SINGULAR.__name__, dn))
    _generic_delete(inst, basedn, log.getChild('_generic_delete'), SINGULAR, dn, args)

def modify(inst, basedn, log, args, warn=True):
    rdn = _get_arg( args.selector, msg="Enter %s to retrieve" % RDN)
    _generic_modify(inst, basedn, log.getChild('_generic_modify'), MANY, rdn, args)

def rename(inst, basedn, log, args, warn=True):
    rdn = _get_arg( args.selector, msg="Enter %s to retrieve" % RDN)
    _generic_rename(inst, basedn, log.getChild('_generic_rename'), MANY, rdn, args)

def create_parser(subparsers):
    ou_parser = subparsers.add_parser('organizationalunit', aliases=['ou'], help='Manage organizational units', formatter_class=CustomHelpFormatter)

    subcommands = ou_parser.add_subparsers(help='action')

    list_parser = subcommands.add_parser('list', help='list', formatter_class=CustomHelpFormatter)
    list_parser.set_defaults(func=list)
    list_parser.add_argument('--full-dn', action='store_true',
                             help="Return the full DN of the entry instead of the RDN value")

    get_parser = subcommands.add_parser('get', help='get', formatter_class=CustomHelpFormatter)
    get_parser.set_defaults(func=get)
    get_parser.add_argument('selector', nargs='?', help='The term to search for')

    get_dn_parser = subcommands.add_parser('get_dn', help='get_dn', formatter_class=CustomHelpFormatter)
    get_dn_parser.set_defaults(func=get_dn)
    get_dn_parser.add_argument('dn', nargs='?', help='The dn to get')

    create_parser = subcommands.add_parser('create', help='create', formatter_class=CustomHelpFormatter)
    create_parser.set_defaults(func=create)
    populate_attr_arguments(create_parser, MUST_ATTRIBUTES)

    delete_parser = subcommands.add_parser('delete', help='deletes the object', formatter_class=CustomHelpFormatter)
    delete_parser.set_defaults(func=delete)
    delete_parser.add_argument('dn', nargs='?', help='The dn to delete')

    modify_parser = subcommands.add_parser('modify', help='modify <add|delete|replace>:<attribute>:<value> ...', formatter_class=CustomHelpFormatter)
    modify_parser.set_defaults(func=modify)
    modify_parser.add_argument('selector', nargs=1, help='The %s to modify' % RDN)
    modify_parser.add_argument('changes', nargs='+', help="A list of changes to apply in format: <add|delete|replace>:<attribute>:<value>")

    rename_parser = subcommands.add_parser('rename', help='rename the object', formatter_class=CustomHelpFormatter)
    rename_parser.set_defaults(func=rename)
    rename_parser.add_argument('selector', help='The %s to rename' % RDN)
    rename_parser.add_argument('new_name', help='A new organizational unit name')
    rename_parser.add_argument('--keep-old-rdn', action='store_true', help="Specify whether the old RDN (i.e. 'ou: old_ou') should be kept as an attribute of the entry or not")



# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4



# --- BEGIN COPYRIGHT BLOCK ---
# Copyright (C) 2019 Red Hat, Inc.
# All rights reserved.
#
# License: GPL (version 3 or any later version).
# See LICENSE for details.
# --- END COPYRIGHT BLOCK ---

import json
from lib389.conflicts import (ConflictEntries, ConflictEntry, GlueEntries, GlueEntry)
from lib389.cli_base import CustomHelpFormatter

conflict_attrs = ['nsds5replconflict', '*']


def list_conflicts(inst, basedn, log, args):
    conflicts = ConflictEntries(inst, args.suffix).list()
    if args.json:
        results = []
        for conflict in conflicts:
            results.append(json.loads(conflict.get_all_attrs_json()))
        log.info(json.dumps({'type': 'list', 'items': results}, indent=4))
    else:
        if len(conflicts) > 0:
            for conflict in conflicts:
                log.info(conflict.display(conflict_attrs))
        else:
            log.info("There were no conflict entries found under the suffix")


def cmp_conflict(inst, basedn, log, args):
    conflict = ConflictEntry(inst, args.DN)
    valid_entry = conflict.get_valid_entry()

    if args.json:
        results = []
        results.append(json.loads(conflict.get_all_attrs_json()))
        results.append(json.loads(valid_entry.get_all_attrs_json()))
        log.info(json.dumps({'type': 'list', 'items': results}, indent=4))
    else:
        log.info("Conflict Entry:\n")
        log.info(conflict.display(conflict_attrs))
        log.info("Valid Entry:\n")
        log.info(valid_entry.display(conflict_attrs))


def del_conflict(inst, basedn, log, args):
    conflict = ConflictEntry(inst, args.DN)
    conflict.delete()


def del_all_conflicts(inst, basedn, log, args):
    count = 0
    failed = 0
    conflicts = ConflictEntries(inst, args.suffix).list()
    for conflict in conflicts:
        try:
            log.info(f"Deleting conflict entry: {conflict.dn}")
            conflict.delete()
            count += 1
        except Exception as e:
            log.error(f"Failed to delete conflict entry {conflict.dn}: {str(e)}")
            failed += 1

    if failed > 0:
        log.warning(f"Deleted {count} conflict entries, {failed} failed.")
    else:
        log.info(f"Deleted {count} conflict entries.")


def swap_conflict(inst, basedn, log, args):
    conflict = ConflictEntry(inst, args.DN)
    conflict.swap()


def convert_conflict(inst, basedn, log, args):
    conflict = ConflictEntry(inst, args.DN)
    conflict.convert(args.new_rdn)


def list_glue(inst, basedn, log, args):
    glues = GlueEntries(inst, args.suffix).list()
    if args.json:
        results = []
        for glue in glues:
            results.append(json.loads(glue.get_all_attrs_json()))
        log.info(json.dumps({'type': 'list', 'items': results}, indent=4))
    else:
        if len(glues) > 0:
            for glue in glues:
                log.info(glue.display(conflict_attrs))
        else:
            log.info("There were no glue entries found under the suffix")


def del_glue(inst, basedn, log, args):
    glue = GlueEntry(inst, args.DN)
    glue.delete_all()


def convert_glue(inst, basedn, log, args):
    glue = GlueEntry(inst, args.DN)
    glue.convert()


def create_parser(subparsers):
    conflict_parser = subparsers.add_parser('repl-conflict', help="Manage replication conflicts", formatter_class=CustomHelpFormatter)
    subcommands = conflict_parser.add_subparsers(help='action')

    # coinflict entry arguments
    list_parser = subcommands.add_parser('list', help="List conflict entries", formatter_class=CustomHelpFormatter)
    list_parser.add_argument('suffix', help='Sets the backend name, or suffix, to look for conflict entries')
    list_parser.set_defaults(func=list_conflicts)

    cmp_parser = subcommands.add_parser('compare', help="Compare the conflict entry with its valid counterpart", formatter_class=CustomHelpFormatter)
    cmp_parser.add_argument('DN', help='The DN of the conflict entry')
    cmp_parser.set_defaults(func=cmp_conflict)

    del_parser = subcommands.add_parser('delete', help="Delete a conflict entry", formatter_class=CustomHelpFormatter)
    del_parser.add_argument('DN', help='The DN of the conflict entry')
    del_parser.set_defaults(func=del_conflict)

    del_all_parser = subcommands.add_parser('delete-all', help="Delete all conflict entries", formatter_class=CustomHelpFormatter)
    del_all_parser.add_argument('suffix', help='Sets the backend name, or suffix, to remove all conflict entries from')
    del_all_parser.set_defaults(func=del_all_conflicts)

    replace_parser = subcommands.add_parser('swap', help="Replace the valid entry with the conflict entry", formatter_class=CustomHelpFormatter)
    replace_parser.add_argument('DN', help='The DN of the conflict entry')
    replace_parser.set_defaults(func=swap_conflict)

    replace_parser = subcommands.add_parser('convert', help="Convert the conflict entry to a valid entry, "
                                                            "while keeping the original valid entry counterpart.  "
                                                            "This requires that the converted conflict entry have "
                                                            "a new RDN value.  For example: \"cn=my_new_rdn_value\".")
    replace_parser.add_argument('DN', help='The DN of the conflict entry')
    replace_parser.add_argument('--new-rdn', required=True, help="Sets the new RDN for the converted conflict entry.  "
                                                                 "For example: \"cn=my_new_rdn_value\"")
    replace_parser.set_defaults(func=convert_conflict)

    # Glue entry arguments
    list_glue_parser = subcommands.add_parser('list-glue', help="List replication glue entries", formatter_class=CustomHelpFormatter)
    list_glue_parser.add_argument('suffix', help='The backend name, or suffix, to look for glue entries')
    list_glue_parser.set_defaults(func=list_glue)

    del_glue_parser = subcommands.add_parser('delete-glue', help="Delete the glue entry and its child entries", formatter_class=CustomHelpFormatter)
    del_glue_parser.add_argument('DN', help='The DN of the glue entry')
    del_glue_parser.set_defaults(func=del_glue)

    convert_glue_parser = subcommands.add_parser('convert-glue', help="Convert the glue entry into a regular entry", formatter_class=CustomHelpFormatter)
    convert_glue_parser.add_argument('DN', help='The DN of the glue entry')
    convert_glue_parser.set_defaults(func=convert_glue)



# --- BEGIN COPYRIGHT BLOCK ---
# Copyright (C) 2017 Red Hat, Inc.
# All rights reserved.
#
# License: GPL (version 3 or any later version).
# See LICENSE for details.
# --- END COPYRIGHT BLOCK ---

# These are the operation runners for configuring a server to look like a certain
# version.

# Generally these are one way, upgrade only.

class baseconfig(object):
    def __init__(self, instance):
        self._instance = instance

    def apply_config(self, install=False, upgrade=False, interactive=False):
        # Go through the list of operations
        # Can we assert the types?
        for op in self._operations:
            op.apply(install, upgrade, interactive)

# This just serves as a base
class configoperation(object):
    def __init__(self, instance):
        self._instance = instance
        self.install = True
        self.upgrade = True
        self.description = None

    def apply(self, install, upgrade, interactive):
        # How do we want to handle interactivity?
        if not ((install and self.install) or (upgrade and self.upgrade)):
            self._instance.debug()
            return False
        if interactive:
            raise Exception('Interaction not yet supported')
        self._apply()

    def _apply(self):
        # The consumer must over-ride this.
        raise Exception('Not implemented!')




# --- BEGIN COPYRIGHT BLOCK ---
# Copyright (C) 2016 Red Hat, Inc.
# All rights reserved.
#
# License: GPL (version 3 or any later version).
# See LICENSE for details.
# --- END COPYRIGHT BLOCK ---
#
from lib389._constants import *
from lib389 import DirSrv, Entry
import pytest
import logging

import ldap
import ldap.sasl

logging.getLogger(__name__).setLevel(logging.DEBUG)
log = logging.getLogger(__name__)

INSTANCE_PORT = 54321
INSTANCE_SERVERID = 'standalone'


class TopologyInstance(object):
    def __init__(self, instance):
        instance.open()
        self.instance = instance


@pytest.fixture(scope="module")
def topology(request):
    # Create the realm
    instance = DirSrv(verbose=False)
    instance.log.debug("Instance allocated")
    args = {SER_PORT: INSTANCE_PORT,
            SER_SERVERID_PROP: INSTANCE_SERVERID}
    instance.allocate(args)
    if instance.exists():
        instance.delete()
    instance.create()
    instance.open()

    def fin():
        if instance.exists():
            instance.delete()
    request.addfinalizer(fin)

    return TopologyInstance(instance)


def test_ldctl(topology):
    """
    """
    # Batch create users
    topology.instance.ldclt.create_users('ou=People,%s' % DEFAULT_SUFFIX, max=1999)
    results = topology.instance.search_s('ou=People,%s' % DEFAULT_SUFFIX, ldap.SCOPE_SUBTREE, filterstr='(objectClass=posixAccount)', attrlist=['uid'])
    assert(len(results) == 1000)

    # Run the load test for a few rounds
    topology.instance.ldclt.bind_loadtest('ou=People,%s' % DEFAULT_SUFFIX, max=1999)

if __name__ == "__main__":
    CURRENT_FILE = os.path.realpath(__file__)
    pytest.main("-s -v %s" % CURRENT_FILE)



# --- BEGIN COPYRIGHT BLOCK ---
# Copyright (C) 2026 Red Hat, Inc.
# All rights reserved.
#
# License: GPL (version 3 or any later version).
# See LICENSE for details.
# --- END COPYRIGHT BLOCK ---

from collections import OrderedDict, namedtuple
import json
import ldap
import os
import sys
from lib389.config import Config, Encryption, RSA, EncryptionModules
from lib389.nss_ssl import NssSsl
from lib389.cert_manager import CertManager
from lib389.cli_base import _warn, CustomHelpFormatter


Props = namedtuple('Props', ['cls', 'attr', 'help', 'values'])

onoff = ('on', 'off')
protocol_versions = ('SSLv3', 'TLS1.0', 'TLSv1.1', 'TLSv1.2', 'TLSv1.3', '')
SECURITY_ATTRS_MAP = OrderedDict([
    ('security', Props(Config, 'nsslapd-security',
                       'Enables or disables security',
                       onoff)),
    ('listen-host', Props(Config, 'nsslapd-securelistenhost',
                          'Sets the host or IP address to listen on for LDAPS',
                          str)),
    ('secure-port', Props(Config, 'nsslapd-securePort',
                          'Sets the port for LDAPS to listen on',
                          range(1, 65536))),
    ('tls-client-auth', Props(Encryption, 'nsSSLClientAuth',
                              'Configures client authentication requirement',
                              ('off', 'allowed', 'required'))),
    ('tls-client-renegotiation', Props(Encryption, 'nsTLSAllowClientRenegotiation',
                                       'Allows client TLS renegotiation',
                                       onoff)),
    ('require-secure-authentication', Props(Config, 'nsslapd-require-secure-binds',
                                            'Configures whether binds over LDAPS, StartTLS, or SASL are required',
                                            onoff)),
    ('check-hostname', Props(Config, 'nsslapd-ssl-check-hostname',
                             'Checks the subject of remote certificate against the hostname',
                             onoff)),
    ('extract-pemfiles', Props(Config, 'nsslapd-extract-pemfiles',
                               'At server shutdown extract the server\'s certificates and keys to PEM files',
                               onoff)),
    ('verify-cert-chain-on-startup', Props(Config, 'nsslapd-validate-cert',
                                           'Validates the server certificate during startup',
                                           ('warn', *onoff))),
    ('session-timeout', Props(Encryption, 'nsSSLSessionTimeout',
                              'Sets the secure session timeout',
                              int)),
    ('tls-protocol-min', Props(Encryption, 'sslVersionMin',
                               'Sets the minimal allowed secure protocol version',
                               protocol_versions)),
    ('tls-protocol-max', Props(Encryption, 'sslVersionMax',
                               'Sets the maximal allowed secure protocol version',
                               protocol_versions)),
    ('allow-insecure-ciphers', Props(Encryption, 'allowWeakCipher',
                                     'Allows weak ciphers for legacy use',
                                     onoff)),
    ('allow-weak-dh-param', Props(Encryption, 'allowWeakDHParam',
                                  'Allows short DH params for legacy use',
                                  onoff)),
    ('cipher-pref', Props(Encryption, 'nsSSL3Ciphers',
                          'Directly sets the nsSSL3Ciphers attribute. It is a comma-separated list '
                          'of cipher names (prefixed with + or -), optionally including +all or -all. The attribute '
                          'may optionally be prefixed by keyword "default". Please refer to documentation of '
                          'the attribute for a more detailed description.',
                          onoff)),
])

RSA_ATTRS_MAP = OrderedDict([
    ('tls-allow-rsa-certificates', Props(RSA, 'nsSSLActivation',
                                         'Activates the use of RSA certificates', onoff)),
    ('nss-cert-name', Props(RSA, 'nsSSLPersonalitySSL',
                            'Sets the server certificate name in NSS DB', str)),
    ('nss-token', Props(RSA, 'nsSSLToken',
                        'Sets the security token name (module of NSS DB)', str))
])


def _security_generic_get(inst, basedn, log, args, attrs_map):
    result = {}
    for attr, props in attrs_map.items():
        val = props.cls(inst).get_attr_val_utf8(props.attr)
        if val is None:
            val = ""
        result[props.attr.lower()] = val
    if args.json:
        log.info(json.dumps({'type': 'list', 'items': result}, indent=4))
    else:
        log.info('\n'.join([f'{attr}: {value or ""}' for attr, value in result.items()]))


def _security_generic_set(inst, basedn, log, args, attrs_map):
    for attr, props in attrs_map.items():
        arg = getattr(args, attr.replace('-', '_'))
        if arg is None:
            continue
        dsobj = props.cls(inst)
        if arg != "":
            dsobj.replace(props.attr, arg)
        else:
            dsobj.remove_all(props.attr)
    log.info("Successfully updated security configuration")


def _security_generic_get_parser(parent, attrs_map, help):
    p = parent.add_parser('get', help=help, formatter_class=CustomHelpFormatter)
    p.set_defaults(func=lambda *args: _security_generic_get(*args, attrs_map))
    return p


def _security_generic_set_parser(parent, attrs_map, help, description):
    p = parent.add_parser('set', help=help, description=description, formatter_class=CustomHelpFormatter)
    p.set_defaults(func=lambda *args: _security_generic_set(*args, attrs_map))
    for opt, params in attrs_map.items():
        p.add_argument(f'--{opt}', help=f'{params[2]} ({params[1]})')
    return p


def _security_ciphers_change(mode, ciphers, inst, log):
    log = log.getChild('_security_ciphers_change')
    if ('default' in ciphers) or ('all' in ciphers):
        raise ValueError("Use ciphers' names only. Keywords 'default' and 'all' are "
                         "ignored. Please, instead specify them manually using 'set' command.")
    enc = Encryption(inst)
    if enc.change_ciphers(mode, ciphers) is False:
        log.error('Setting new ciphers failed.')


def _security_generic_toggle(inst, basedn, log, args, cls, attr, value, thing):
    cls(inst).set(attr, value)


def _security_generic_toggle_parsers(parent, cls, attr, help_pattern):
    def add_parser(action, value):
        p = parent.add_parser(action.lower(), help=help_pattern.format(action), formatter_class=CustomHelpFormatter)
        p.set_defaults(func=lambda *args: _security_generic_toggle(*args, cls, attr, value, action))
        return p

    return list(map(add_parser, ('Enable', 'Disable'), ('on', 'off')))


def _resolve_pkcs12_password(args):
    if args.pkcs12_pin_text:
        return args.pkcs12_pin_text

    if args.pkcs12_pin_stdin:
        return sys.stdin.readline().rstrip("\n")

    if args.pkcs12_pin_path:
        with open(args.pkcs12_pin_path) as f:
            return f.read().rstrip("\n")

    return None

def _dump_cert(cert, json_output: bool = False, log = None):
    """
    Print or return a certificate's details in text or JSON format.

    :param cert: dict describing a certificate
    :param json_output: If True print JSON else print text
    :param log: Optional logger to output text
    """
    if not isinstance(cert, dict):
        raise TypeError(f"Expected dict, got {type(cert)}")

    if json_output:
        return {
            "type": "certificate",
            "attrs": {
                "nickname": cert["cn"],
                "subject": cert["subject"],
                "issuer": cert["issuer"],
                "expires": cert["expires"],
                "flags": cert["trust_flags"],
            }
        }
    else:
        msg = (
            f"Certificate Name: {cert['cn']}\n"
            f"Subject DN: {cert['subject']}\n"
            f"Issuer DN: {cert['issuer']}\n"
            f"Expires: {cert['expires']}\n"
            f"Trust Flags: {cert['trust_flags']}\n"
        )
        if log:
            log.info(msg)

def security_enable(inst, basedn, log, args):
    dbpath = inst.get_cert_dir()
    tlsdb = NssSsl(dbpath=dbpath)
    certs = tlsdb.list_certs()
    if len(certs) == 0:
        raise ValueError('There are no server certificates in the security ' +
                         'database, security can not be enabled.')

    if len(certs) == 1:
        # If there is only cert make sure it is set as the server certificate
        RSA(inst).set('nsSSLPersonalitySSL', certs[0]['cn'])
    elif args.cert_name is not None:
        # A certificate nickname was provided, set it as the server certificate
        RSA(inst).set('nsSSLPersonalitySSL', args.cert_name)

    # it should now be safe to enable security
    Config(inst).set('nsslapd-security', 'on')
    log.info("Successfully enabled security")


def security_disable(inst, basedn, log, args):
    Config(inst).set('nsslapd-security', 'off')


def security_ciphers_enable(inst, basedn, log, args):
    _security_ciphers_change('+', args.cipher, inst, log)


def security_ciphers_disable(inst, basedn, log, args):
    _security_ciphers_change('-', args.cipher, inst, log)


def security_ciphers_set(inst, basedn, log, args):
    enc = Encryption(inst)
    enc.ciphers = args.cipher_string.lstrip().split(',')


def security_ciphers_get(inst, basedn, log, args):
    enc = Encryption(inst)
    if args.json:
        log.info({'type': 'list', 'items': enc.ciphers})
    else:
        val = ','.join(enc.ciphers)
        log.info(val if val != '' else '<undefined>')


def security_ciphers_list(inst, basedn, log, args):
    enc = Encryption(inst)

    if args.enabled:
        lst = enc.enabled_ciphers
    elif args.supported:
        lst = enc.supported_ciphers
    elif args.disabled:
        lst = set(enc.supported_ciphers) - set(enc.enabled_ciphers)
    else:
        lst = enc.ciphers

    if args.json:
        log.info(json.dumps({'type': 'list', 'items': lst}, indent=4))
    else:
        if lst == []:
            log.getChild('security').warn('List of ciphers is empty')
        else:
            for item in lst:
                log.info(item)


def security_disable_plaintext_port(inst, basedn, log, args, warn=True):
    if warn and args.json is False:
        _warn(True, msg="Disabling plaintext ldap port - you must have ldaps configured")
    inst.config.disable_plaintext_port()
    log.info("Plaintext port disabled - please restart your instance to take effect")
    log.info("To undo this change run the subcommand - 'dsconf <instance> config replace nsslapd-port=<port number>'")


def cert_add(inst, basedn, log, args):
    """Add server certificate
    """
    # Verify file and certificate name
    if not os.path.isfile(args.file):
        raise ValueError(f'Certificate file "{args.file}" does not exist')

    pkcs12_password = None
    pkcs12_file = args.file.lower().endswith((".p12", ".pfx"))
    if pkcs12_file:
        pkcs12_password = _resolve_pkcs12_password(args)

    certmgr = CertManager(instance=inst)
    cert = certmgr.get_cert(args.name)
    if cert:
        raise ValueError(f"Certificate '{args.name}' already exists")

    certmgr.add_cert(
        args.file,
        args.name,
        pkcs12_password=pkcs12_password,
        primary=args.primary_cert,
        ca=False,
        force=args.force
    )
    log.info("Successfully added certificate")

def cacert_add(inst, basedn, log, args):
    """Add CA certificate, or CA certificate bundle
    """
    # Verify file and certificate name
    if not os.path.isfile(args.file):
        raise ValueError(f'Certificate file "{args.file}" does not exist')

    # Does it make sense to add a CA cert from p12 container ?
    if args.file.lower().endswith((".p12", ".pfx")):
        raise ValueError("PKCS#12 CA certificates not supported. Use PEM or DER file")

    certmgr = CertManager(instance=inst)
    certmgr.add_ca_cert(args.file, args.name, force=args.force)
    log.info("Successfully added CA certificate")

def cert_list(inst, basedn, log, args):
    """List all the server certificates
    """
    certmgr = CertManager(instance=inst)
    certs = certmgr.list_certs()
    if not certs:
        if args.json:
            log.info(json.dumps([], indent=4))
        else:
            log.info("No certificates found.")
        return

    if args.json:
        output = [_dump_cert(cert, json_output=True) for cert in certs]
        log.info(json.dumps(output, indent=4))
    else:
        for cert in certs:
            _dump_cert(cert, json_output=False, log=log)

def cacert_list(inst, basedn, log, args):
    """List all CA certs
    """
    certmgr = CertManager(instance=inst)
    ca_certs = certmgr.list_ca_certs()
    if not ca_certs:
        if args.json:
            log.info(json.dumps([], indent=4))
        else:
            log.info("No CA certificates found.")
        return

    if args.json:
        output = [_dump_cert(cert, json_output=True) for cert in ca_certs]
        log.info(json.dumps(output, indent=4))
    else:
        for cert in ca_certs:
            _dump_cert(cert, json_output=False, log=log)

def cert_get(inst, basedn, log, args):
    """Get the details about a server certificate
    """
    certmgr = CertManager(instance=inst)
    cert = certmgr.get_cert(args.name)
    if not cert:
        raise ValueError(f"Certificate '{args.name}' not found.")

    if "C" in cert.get("trust_flags", ""):
        return

    if args.json:
        output = _dump_cert(cert, json_output=args.json)
        log.info(json.dumps(output, indent=4))
    else:
        _dump_cert(cert, json_output=args.json, log=log)

def cacert_get(inst, basedn, log, args):
    """Get the details about a CA certificate
    """
    certmgr = CertManager(instance=inst)
    cert = certmgr.get_cert(args.name)
    if not cert:
        raise ValueError(f"Certificate '{args.name}' not found.")

    if "C" not in cert.get("trust_flags", ""):
        return

    if args.json:
        output = _dump_cert(cert, json_output=args.json)
        log.info(json.dumps(output, indent=4))
    else:
        _dump_cert(cert, json_output=args.json, log=log)

def csr_list(inst, basedn, log, args):
    """
    List all files with .csr extension in instance config dir
    """
    csr_list = []
    tlsdb = NssSsl(dirsrv=inst)
    details = tlsdb._csr_list(args.path)

    for detail in details:
        if args.json:
            csr_list.append(
                {
                    "type": "csr",
                    "attrs": {
                        'modified': detail[0],
                        'subject': detail[1],
                        'subject_alt_names': detail[2],
                        'name': detail[3],
                    }
                }
            )
        else:
            log.info(f'Modified: {detail[0]}')
            log.info(f'Subject: {detail[1]}')
            log.info(f'Subject Alternative Names: {", ".join(detail[2])}')
            log.info(f'Name: {detail[3]}\n')

    if args.json:
        log.info(json.dumps(csr_list, indent=4))


def csr_get(inst, basedn, log, args):
    """
    Show PEM format of a csr
    """
    tlsdb = NssSsl(dirsrv=inst)
    details = tlsdb._csr_show(args.name)
    log.info(f"{details}")


def csr_gen(inst, basedn, log, args):
    """
    Generate a .csr file in instance config dir
    """
    tlsdb = NssSsl(dirsrv=inst)
    alt_names = args.alt_names
    subject = args.subject
    name = args.name
    out_path = tlsdb.create_rsa_key_and_csr(alt_names, subject, name)
    log.info(out_path)


def csr_del(inst, basedn, log, args):
    """
    Delete a .csr file from instance config dir
    """
    csr_dir = inst.get_cert_dir()
    file_path = os.path.join(csr_dir, args.name + ".csr")
    try:
        os.remove(file_path)
    except FileNotFoundError:
        raise ValueError(file_path + " not found")
    log.info(f"Successfully deleted: " + file_path)


def cert_edit(inst, basedn, log, args):
    """Edit cert
    """
    certmgr = CertManager(instance=inst)
    certmgr.edit_cert_trust(args.name, args.flags)
    log.info("Successfully edited certificate trust flags")


def cert_del(inst, basedn, log, args):
    """Delete cert
    """
    certmgr = CertManager(instance=inst)
    try:
        certmgr.del_cert(args.name)
        log.info(f"Successfully deleted certificate")
    except ValueError as e:
        log.error(f"Failed to delete certificate '{args.name}': {e}")

def key_list(inst, basedn, log, args):
    """
    List keys in the NSS DB
    """
    key_list = []
    tls = NssSsl(dirsrv=inst)
    keys = tls.list_keys(args.orphan)

    for key in keys:
        if args.json:
            key_list.append(
                {
                    "type": "key",
                    "attrs": {
                                'cipher': key[0],
                                'key_id': key[1],
                                'state': key[2],
                            }
                }
            )
        else:
            log.info('Cipher: {}'.format(key[0]))
            log.info('Key Id: {}'.format(key[1]))
            log.info('State: {}\n'.format(key[2]))

    if args.json:
            log.info(json.dumps(key_list, indent=4))


def key_del(inst, basedn, log, args):
    """
    Delete a key from NSS DB
    """
    tls = NssSsl(dirsrv=inst)
    keys = tls.del_key(args.key_id)
    log.info(keys)


def export_cert(inst, basedn, log, args):
    tls = NssSsl(dirsrv=inst)
    nickname = args.nickname
    der_format = False
    output_file = None
    if args.binary_format:
        der_format = True
    if args.output_file is not None:
        output_file = args.output_file
    tls.export_cert(nickname, output_file, der_format)


def encryption_module_add(inst, basedn, log, args):
    """Add an encryption module
    """
    properties = {
        'cn': args.name,
        'nsSSLActivation': "on" if args.activated else "off",
        'nsSSLPersonalitySSL': args.cert_nickname,
        'nsSSLToken': args.token
    }
    if args.server_key_extract_file is not None:
        properties['ServerKeyExtractFile'] = args.server_key_extract_file
    if args.server_cert_extract_file is not None:
        properties['ServerCertExtractFile'] = args.server_cert_extract_file
    encryption_modules = EncryptionModules(instance=inst)
    encryption_modules.create(properties=properties)
    log.info("Successfully added encryption module")


def encryption_module_delete(inst, basedn, log, args):
    """Delete an encryption module
    """
    if args.name.lower() == "rsa":
        # Currently RSA is hardcoded into attribute encryption so it can not
        # be deleted at this time
        raise ValueError("Deletion of RSA encryption module is not allowed")

    try:
        encryption_module = EncryptionModules(instance=inst).get(args.name)
    except ldap.NO_SUCH_OBJECT as e:
        raise ValueError(f"Encryption module '{args.name}' not found")

    encryption_module.delete()
    log.info("Successfully deleted encryption module")


def encryption_module_edit(inst, basedn, log, args):
    """Edit an encryption module
    """
    try:
        encryption_module = EncryptionModules(instance=inst).get(args.name)
    except ldap.NO_SUCH_OBJECT as e:
        raise ValueError(f"Encryption module '{args.name}' not found")

    if args.activate and args.deactivate:
        raise ValueError("Cannot activate and deactivate an encryption module at the same time")

    replace_list = []
    if args.cert_nickname is not None:
        replace_list.append(['nsSSLPersonalitySSL', args.cert_nickname])
    if args.activate:
        replace_list.append(['nsSSLActivation', 'on'])
    if args.deactivate:
        replace_list.append(['nsSSLActivation', 'off'])
    if args.token is not None:
        replace_list.append(['nsSSLToken', args.token])
    if args.server_key_extract_file is not None:
        replace_list.append(['ServerKeyExtractFile', args.server_key_extract_file])
    if args.server_cert_extract_file is not None:
        replace_list.append(['ServerCertExtractFile', args.server_cert_extract_file])

    if len(replace_list) > 0:
        encryption_module.replace_many(*replace_list)
    else:
        raise ValueError("There are no changes to set for this encryption module")

    log.info("Successfully updated encryption module")


def encryption_module_list(inst, basedn, log, args):
    """List encryption module names
    """
    encryption_modules = EncryptionModules(instance=inst)
    encryption_modules_list = encryption_modules.list()

    if args.json:
        entry_list = []
        for encryption_module in encryption_modules_list:
            if args.just_names:
                entry_list.append(encryption_module.get_attr_val_utf8('cn'))
            else:
                entry_list.append(json.loads(encryption_module.get_all_attrs_json()))

        log.info(json.dumps({"type": "list", "items": entry_list}, indent=4))
    else:
        for encryption_module in encryption_modules_list:
            if args.just_names:
                log.info(encryption_module.get_attr_val_utf8('cn'))
            else:
                entry = encryption_module.display()
                updated_entry = entry[:-1]  # remove \n
                log.info(updated_entry)


def encryption_module_get(inst, basedn, log, args):
    """Get an encryption module
    """
    try:
        encryption_module = EncryptionModules(instance=inst).get(args.name)
    except ldap.NO_SUCH_OBJECT as e:
        raise ValueError(f"Encryption module '{args.name}' not found")

    if args.json:
        entry = encryption_module.get_all_attrs_json()
        entry_dict = json.loads(entry)
        log.info(json.dumps(entry_dict, indent=4))
    else:
        entry = encryption_module.display()
        updated_entry = entry[:-1]  # remove \n
        log.info(updated_entry)


def create_parser(subparsers):
    security = subparsers.add_parser('security', help='Manage security settings', formatter_class=CustomHelpFormatter)
    security_sub = security.add_subparsers(help='security')

    # Core security management
    _security_generic_set_parser(security_sub, SECURITY_ATTRS_MAP, 'Set general security options',
        ('Use this command for setting security related options located in cn=config and cn=encryption,cn=config.'
         '\n\nTo enable/disable security you can use enable and disable commands instead.'))
    _security_generic_get_parser(security_sub, SECURITY_ATTRS_MAP, 'Display general security options')
    security_enable_p = security_sub.add_parser('enable', help='Enable security', description=(
        'If missing, create security database, then turn on security functionality. Please note this is usually not'
        ' enough for TLS connections to work - proper setup of CA and server certificate is necessary.'))
    security_enable_p.add_argument('--cert-name', default=None,
        help='Sets the name of the certificate the server should use')
    security_enable_p.set_defaults(func=security_enable)
    security_disable_p = security_sub.add_parser('disable', help='Disable security', description=(
        'Turn off security functionality. The rest of the configuration will be left untouched.'))
    security_disable_p.set_defaults(func=security_disable)

    security_disable_plain_parser = security_sub.add_parser('disable_plain_port',
        help="Disables the plain text LDAP port, allowing only LDAPS to function")
    security_disable_plain_parser.set_defaults(func=security_disable_plaintext_port)

    # Server certificate management
    certs = security_sub.add_parser('certificate', help='Manage TLS certificates', formatter_class=CustomHelpFormatter)
    certs_sub = certs.add_subparsers(help='certificate')
    cert_add_parser = certs_sub.add_parser('add', help='Add a server certificate', description=(
        'Add a server certificate to the NSS database'))
    cert_add_parser.add_argument('--file', required=True,
        help='Sets the file name of the certificate')
    cert_add_parser.add_argument('--name', required=True,
        help='Sets the name/nickname of the certificate')
    cert_add_parser.add_argument('--primary-cert', action='store_true',
                                 help="Sets this certificate as the server's certificate")
    cert_add_parser.add_argument('--pkcs12-pin-text', help='The PKCS#12 password as plain text. WARNING: Password may appear' \
        ' in process list or shell history. Use --pkcs12-pin-stdin or --pkcs12-pin-path to prevent password exposure.')
    cert_add_parser.add_argument('--pkcs12-pin-stdin', help='Read the PKCS#12 password from stdin', action='store_true')
    cert_add_parser.add_argument('--pkcs12-pin-path',  help='Path to a file containing the PKCS#12 password')
    cert_add_parser.add_argument('--do-it', dest="force", help="Force the addition of a certificate that cannot be verified",action='store_true', default=False)
    cert_add_parser.set_defaults(func=cert_add)

    cert_edit_parser = certs_sub.add_parser('set-trust-flags', help='Set the Trust flags',
        description=('Change the trust flags of a server certificate'))
    cert_edit_parser.add_argument('name', help='The name/nickname of the certificate')
    cert_edit_parser.add_argument('--flags', required=True,
        help='Sets the trust flags for the server certificate')
    cert_edit_parser.set_defaults(func=cert_edit)

    cert_del_parser = certs_sub.add_parser('del', help='Delete a certificate',
        description=('Delete a server certificate from the NSS database or DynamicCerts backend.'))
    cert_del_parser.add_argument('name', help='The name/nickname of the certificate')
    cert_del_parser.set_defaults(func=cert_del)

    cert_get_parser = certs_sub.add_parser('get', help="Display a server certificate's information",
        description=('Displays detailed information about a certificate, such as trust attributes, expiration dates, Subject and Issuer DNs '))
    cert_get_parser.add_argument('name', help='Set the name/nickname of the certificate')
    cert_get_parser.set_defaults(func=cert_get)

    cert_list_parser = certs_sub.add_parser('list', help='List the server certificates',
        description=('List all server certificates in the NSS database or DynamicCerts backend.'))
    cert_list_parser.set_defaults(func=cert_list)

    # CA certificate management
    cacerts = security_sub.add_parser('ca-certificate', help='Manage TLS certificate authorities', formatter_class=CustomHelpFormatter)
    cacerts_sub = cacerts.add_subparsers(help='ca-certificate')
    cacert_add_parser = cacerts_sub.add_parser('add', help='Add a Certificate Authority', description=(
        'Add a CA certificate (PEM or DER only) to the NSS database or DynamicCerts backend.'))
    cacert_add_parser.add_argument('--file', required=True,
        help='Path to the CA certificate file (PEM or DER). If adding a PEM bundle then specify multiple names one for '
             'each certificate, otherwise a number increment will be added to the previous name.')
    cacert_add_parser.add_argument('--name', nargs='+', required=True,
        help='Sets the name/nickname of the CA certificate')
    cacert_add_parser.add_argument('--do-it', dest="force", help="Force the addition of a certificate that cannot be verified",action='store_true', default=False)
    cacert_add_parser.set_defaults(func=cacert_add)

    cacert_edit_parser = cacerts_sub.add_parser('set-trust-flags', help='Set the Trust flags',
        description=('Change the trust attributes of a CA certificate.  Certificate Authorities typically use "CT,,"'))
    cacert_edit_parser.add_argument('name', help='The name/nickname of the CA certificate')
    cacert_edit_parser.add_argument('--flags', required=True,
        help='Sets the trust flags for the CA certificate')
    cacert_edit_parser.set_defaults(func=cert_edit)

    cacert_del_parser = cacerts_sub.add_parser('del', help='Delete a certificate',
        description=('Delete a CA certificate from the NSS database'))
    cacert_del_parser.add_argument('name', help='The name/nickname of the CA certificate')
    cacert_del_parser.set_defaults(func=cert_del)

    cacert_get_parser = cacerts_sub.add_parser('get', help="Displays a Certificate Authority's information",
        description=('Get detailed information about a CA certificate, like trust attributes, expiration dates, Subject and Issuer DN'))
    cacert_get_parser.add_argument('name', help='The name/nickname of the CA certificate')
    cacert_get_parser.set_defaults(func=cacert_get)

    cacert_list_parser = cacerts_sub.add_parser('list', help='List the Certificate Authorities',
        description=('List the CA certificates in the NSS database'))
    cacert_list_parser.set_defaults(func=cacert_list)

    # RSA management
    rsa = security_sub.add_parser('rsa', help='Query and update RSA security options', formatter_class=CustomHelpFormatter)
    rsa_sub = rsa.add_subparsers(help='rsa')
    _security_generic_set_parser(rsa_sub, RSA_ATTRS_MAP, 'Set RSA security options',
        ('Use this command for setting RSA (private key) related options located in cn=RSA,cn=encryption,cn=config.'
         '\n\nTo enable/disable RSA you can use enable and disable commands instead.'))
    _security_generic_get_parser(rsa_sub, RSA_ATTRS_MAP, 'Get RSA security options')
    _security_generic_toggle_parsers(rsa_sub, RSA, 'nsSSLActivation', '{} RSA')

    # Encryption module management (Should replace RSA at some point)
    encryption_module = security_sub.add_parser('encryption-module', help='Manage encryption modules',
                                                formatter_class=CustomHelpFormatter)
    encryption_module_sub = encryption_module.add_subparsers(help='encryption-module')
    # Add an encryption module
    encryption_module_add_parser = encryption_module_sub.add_parser('add', help='Add an encryption module',
            description=('Add a new encryption module to the encryption module list.'))
    encryption_module_add_parser.add_argument('name', help='The name of the encryption module')
    encryption_module_add_parser.add_argument('--cert-nickname', help='The personality or nickname of the server certificate',
                                              required=True)
    encryption_module_add_parser.add_argument('--activated', action='store_true',
                                              help='Activate the encryption module.')
    encryption_module_add_parser.add_argument('--token', help='The token of the encryption module. Default is "internal (software)".',
                                              default='internal (software)')
    encryption_module_add_parser.add_argument('--server-key-extract-file', help='The file name of the server key extract file')
    encryption_module_add_parser.add_argument('--server-cert-extract-file', help='The file name of the server cert extract file')
    encryption_module_add_parser.set_defaults(func=encryption_module_add)
    # delete an encryption module
    encryption_module_delete_parser = encryption_module_sub.add_parser('delete', help='Delete an encryption module',
        description=('Delete an encryption module from the encryption module list.'))
    encryption_module_delete_parser.add_argument('name', help='The name of the encryption module')
    encryption_module_delete_parser.set_defaults(func=encryption_module_delete)
    # edit an encryption module
    encryption_module_edit_parser = encryption_module_sub.add_parser('edit', help='Edit an encryption module',
        description=('Edit an encryption module in the encryption module list.'))
    encryption_module_edit_parser.add_argument('name', help='The name of the encryption module')
    encryption_module_edit_parser.add_argument('--cert-nickname', help='The personality or nickname of the server certificate')
    encryption_module_edit_parser.add_argument('--activate', action='store_true', help='Activate the encryption module')
    encryption_module_edit_parser.add_argument('--deactivate', action='store_true', help='Deactivate the encryption module')
    encryption_module_edit_parser.add_argument('--token', help='The token of the encryption module.')
    encryption_module_edit_parser.add_argument('--server-key-extract-file', help='The file name of the server key extract file')
    encryption_module_edit_parser.add_argument('--server-cert-extract-file', help='The file name of the server cert extract file')
    encryption_module_edit_parser.set_defaults(func=encryption_module_edit)
    # list encryption modules
    encryption_module_list_parser = encryption_module_sub.add_parser('list', help='List encryption modules',
        description=('List all encryption modules in the encryption module list.'))
    encryption_module_list_parser.add_argument('--just-names', action='store_true',
                                               help='Just list the modules by its name and not the full entry')
    encryption_module_list_parser.set_defaults(func=encryption_module_list)
    # Get an encryption module
    encryption_module_get_parser = encryption_module_sub.add_parser('get', help='Get an encryption module',
        description=('Get an encryption module from the encryption module list.'))
    encryption_module_get_parser.add_argument('name', help='The name of the encryption module')
    encryption_module_get_parser.set_defaults(func=encryption_module_get)

    # Cipher management
    ciphers = security_sub.add_parser('ciphers', help='Manage secure ciphers', formatter_class=CustomHelpFormatter)
    ciphers_sub = ciphers.add_subparsers(help='ciphers')

    ciphers_enable = ciphers_sub.add_parser('enable', help='Enable ciphers', description=(
        'Use this command to enable specific ciphers.'))
    ciphers_enable.set_defaults(func=security_ciphers_enable)
    ciphers_enable.add_argument('cipher', nargs='+')

    ciphers_disable = ciphers_sub.add_parser('disable', help='Disable ciphers', description=(
        'Use this command to disable specific ciphers.'))
    ciphers_disable.set_defaults(func=security_ciphers_disable)
    ciphers_disable.add_argument('cipher', nargs='+')

    ciphers_get = ciphers_sub.add_parser('get', help='Get ciphers attribute', description=(
        'Use this command to get contents of nsSSL3Ciphers attribute.'))
    ciphers_get.set_defaults(func=security_ciphers_get)

    ciphers_set = ciphers_sub.add_parser('set', help='Set ciphers attribute', description=(
        'Use this command to directly set nsSSL3Ciphers attribute. It is a comma separated list '
        'of cipher names (prefixed with + or -), optionally including +all or -all. The attribute '
        'may optionally be set to keyword default. Please refer to documentation of '
        'the attribute for a more detailed description.'))
    ciphers_set.set_defaults(func=security_ciphers_set)
    ciphers_set.add_argument('cipher_string', metavar='cipher-string')

    ciphers_list = ciphers_sub.add_parser('list', help='List ciphers', description=(
        'List secure ciphers. Without arguments, list ciphers as configured in nsSSL3Ciphers attribute.'))
    ciphers_list.set_defaults(func=security_ciphers_list)
    ciphers_list_group = ciphers_list.add_mutually_exclusive_group()
    ciphers_list_group.add_argument('--enabled', action='store_true',
                                    help='Lists only enabled ciphers')
    ciphers_list_group.add_argument('--supported', action='store_true',
                                    help='Lists only supported ciphers')
    ciphers_list_group.add_argument('--disabled', action='store_true',
                                    help='Lists only supported ciphers but without enabled ciphers')

    # Certificate Signing Request Management
    csr = security_sub.add_parser('csr', help='Manage certificate signing requests', formatter_class=CustomHelpFormatter)
    csr_sub = csr.add_subparsers(help='csr')

    csr_list_parser = csr_sub.add_parser('list', help='List CSRs', description=('List all CSR files in instance'
        ' configuration directiory'))
    csr_list_parser.add_argument('--path', '-p', default=None, help="Directory contanining CSR file")
    csr_list_parser.set_defaults(func=csr_list)

    csr_get_parser = csr_sub.add_parser('get', help='Display CSR content', description=('Displays the contents of a CSR, '
        ' which can be used for submittal to CA'))
    csr_get_parser.add_argument('name', help="Name of the CSR file to display")
    csr_get_parser.set_defaults(func=csr_get)

    csr_req_parser = csr_sub.add_parser('req', help='Generate a Certificate Signing Request',
        description=('Generate a CSR that can be submitted to a CA for verification'))
    csr_req_parser.add_argument('--subject', '-s', required=True, help="Subject field")
    csr_req_parser.add_argument('--name', '-n', required=True, help="Name")
    csr_req_parser.add_argument('alt_names', nargs='*',
         help="CSR alternative names. These are auto-detected if not provided")
    csr_req_parser.set_defaults(func=csr_gen)

    csr_delete_parser = csr_sub.add_parser('del', help='Delete a CSR file', description=('Delete a CSR file'), formatter_class=CustomHelpFormatter)
    csr_delete_parser.add_argument('name', help="Name of the CSR file to delete")
    csr_delete_parser.set_defaults(func=csr_del)

    # Key Management
    key = security_sub.add_parser('key', help='Manage keys in NSS DB', formatter_class=CustomHelpFormatter)
    key_sub = key.add_subparsers(help='key')

    key_list_parser = key_sub.add_parser('list', help='List all keys in NSS DB', formatter_class=CustomHelpFormatter)
    key_list_parser.add_argument('--orphan', action='store_true', help='List orphan keys (An orphan key is'
        ' a private key in the NSS DB for which there is NO cert with the corresponding '
        ' public key). An orphan key is created during CSR generation, when the associated certificate is imported'
        ' into the NSS DB, its orphan state will be removed.')
    key_list_parser.set_defaults(func=key_list)

    key_del_parser = key_sub.add_parser('del', help='Delete a key from NSS DB', description=('Remove a'
        ' key from the NSS DB. Make sure the key is not in use before you delete'))
    key_del_parser.add_argument('key_id', help='This is the key ID displayed when listing keys')
    key_del_parser.set_defaults(func=key_del)

    # Export certificate
    export_cert_parser = security_sub.add_parser(
        'export-cert',
        help="Export a certificate to PEM or DER/Binary format.  PEM format is the default"
    )
    export_cert_parser.add_argument('nickname', help="The name of the certificate to export")
    export_cert_parser.add_argument('--binary-format', action='store_true',
                                    help="Export certificate in DER/binary format")
    export_cert_parser.add_argument('--output-file',
                                    help='The name for the exported certificate.  Default name is the certificate '
                                         'nickname with an extension of ".pem" or ".crt"')
    export_cert_parser.set_defaults(func=export_cert)



# --- BEGIN COPYRIGHT BLOCK ---
# Copyright (C) 2019 William Brown <william@blackhats.net.au>
# Copyright (C) 2025 Red Hat, Inc.
# All rights reserved.
#
# License: GPL (version 3 or any later version).
# See LICENSE for details.
# --- END COPYRIGHT BLOCK ---

import datetime
import json
import os
from lib389._constants import (DB_IMPL_BDB, DB_IMPL_MDB)
from lib389.monitor import (Monitor, MonitorLDBM, MonitorSNMP, MonitorDiskSpace, MonitorMemberOf)
from lib389.chaining import (ChainingLinks)
from lib389.backend import Backends
from lib389.utils import convert_bytes
from lib389.cli_base import (_format_status, _format_status_with_option,
                             CustomHelpFormatter)


def monitor(inst, basedn, log, args):
    """Server monitor"""
    server_monitor = Monitor(inst)
    _format_status_with_option(log, server_monitor, args.json,
                               args.just_resources)


def backend_monitor(inst, basedn, log, args):
    bes = Backends(inst)
    if args.backend:
        be = bes.get(args.backend)
        be_monitor = be.get_monitor()
        _format_status(log, be_monitor, args.json)
    else:
        for be in bes.list():
            be_monitor = be.get_monitor()
            _format_status(log, be_monitor, args.json)


def ldbm_monitor(inst, basedn, log, args):
    ldbm_monitor = MonitorLDBM(inst)
    _format_status(log, ldbm_monitor, args.json)


def snmp_monitor(inst, basedn, log, args):
    snmp_monitor = MonitorSNMP(inst)
    _format_status(log, snmp_monitor, args.json)


def chaining_monitor(inst, basedn, log, args):
    links = ChainingLinks(inst)
    if args.backend:
        link = links.get(args.backend)
        link_monitor = link.get_monitor()
        _format_status(log, link_monitor, args.json)
    else:
        for link in links.list():
            link_monitor = link.get_monitor()
            _format_status(log, link_monitor, args.json)
            # Inject a new line for now ... see https://github.com/389ds/389-ds-base/issues/3248
            log.info("")


def disk_monitor(inst, basedn, log, args):
    disk_space_mon = MonitorDiskSpace(inst)
    disks = disk_space_mon.get_disks()
    disk_list = []
    for disk in disks:
        # partition="/" size="52576092160" used="25305038848" available="27271053312" use%="48"
        parts = disk.split()
        mount = parts[0].split('=')[1].strip('"')
        disk_size = convert_bytes(parts[1].split('=')[1].strip('"'))
        used = convert_bytes(parts[2].split('=')[1].strip('"'))
        avail = convert_bytes(parts[3].split('=')[1].strip('"'))
        percent = parts[4].split('=')[1].strip('"')
        if args.json:
            disk_list.append({
                'mount': mount,
                'size': disk_size,
                'used': used,
                'avail': avail,
                'percent': percent
            })
        else:
            log.info("Partition: " + mount)
            log.info("Size: " + disk_size)
            log.info("Used Space: " + used)
            log.info("Available Space: " + avail)
            log.info("Percentage Used: " + percent + "%\n")

    if args.json:
        log.info(json.dumps({"type": "list", "items": disk_list}, indent=4))


def db_monitor(inst, basedn, log, args):
    """Report on all the database statistics
    """
    ldbm_monitor = MonitorLDBM(inst)
    backends_obj = Backends(inst)
    backend_objs = []
    args_backends = None

    # Gather all the backends
    if args.backends is not None:
        # This is a space separated list, it could be backend names or suffixes
        args_backends = args.backends.lower().split()

    for be in backends_obj.list():
        if args_backends is not None:
            for arg_be in args_backends:
                if '=' in arg_be:
                    # We have a suffix
                    if arg_be == be.get_suffix():
                        backend_objs.append(be)
                        break
                else:
                    # We have a backend name
                    if arg_be == be.rdn.lower():
                        backend_objs.append(be)
                        break
        else:
            # Get all the backends
            backend_objs.append(be)

    if args_backends is not None and len(backend_objs) == 0:
        raise ValueError("Could not find any backends from the provided list: {}".format(args.backends))

    # Gather the global DB stats
    report_time = datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S")
    ldbm_mon = ldbm_monitor.get_status()
    ndn_cache_enabled = inst.config.get_attr_val_utf8('nsslapd-ndn-cache-enabled') == 'on'

    # Build global cache stats
    result = {
        'date': report_time,
        'backends': {},
    }

    if ldbm_monitor.inst_db_impl == DB_IMPL_BDB:
        dbcachesize = int(ldbm_mon['nsslapd-db-cache-size-bytes'][0])
        # Warning: there are two different page sizes associated with bdb:
        # - nsslapd-db-mp-pagesize the db mempool (i.e the db cache) page size which is usually 4K
        # - nsslapd-db-pagesize the db instances (i.e id2entry, indexes, changelog) page size which
        #   is usually 8K
        # To compute the db cache statistics we must use the nsslapd-db-mp-pagesize
        if 'nsslapd-db-mp-pagesize' in ldbm_mon:
            pagesize = int(ldbm_mon['nsslapd-db-mp-pagesize'][0])
        else:
            # targeting a remote instance that does not have github issue 5550 fix.
            # So lets use the usual default file system preferred block size
            # db cache free statistics may be wrong but we gave no way to
            # compute it rightly.
            pagesize = 4096

        dbhitratio = ldbm_mon['dbcachehitratio'][0]
        dbcachepagein = ldbm_mon['dbcachepagein'][0]
        dbcachepageout = ldbm_mon['dbcachepageout'][0]
        dbroevict = ldbm_mon['nsslapd-db-page-ro-evict-rate'][0]
        dbpages = int(ldbm_mon['nsslapd-db-pages-in-use'][0])
        dbcachefree = max(int(dbcachesize - (pagesize * dbpages)), 0)
        dbcachefreeratio = dbcachefree/dbcachesize

        result['dbcache'] = {
                'hit_ratio': dbhitratio,
                'free': convert_bytes(str(dbcachefree)),
                'free_percentage': "{:.1f}".format(dbcachefreeratio * 100),
                'roevicts': dbroevict,
                'pagein': dbcachepagein,
                'pageout': dbcachepageout
        }

    # Add NDN cache stats only if enabled
    if ndn_cache_enabled:
        try:
            ndnratio = ldbm_mon['normalizeddncachehitratio'][0]
            ndncursize = int(ldbm_mon['currentnormalizeddncachesize'][0])
            ndnmaxsize = int(ldbm_mon['maxnormalizeddncachesize'][0])
            ndncount = ldbm_mon['currentnormalizeddncachecount'][0]
            ndnevictions = ldbm_mon['normalizeddncacheevictions'][0]
            if ndncursize > ndnmaxsize:
                ndnfree = 0
                ndnfreeratio = 0
            else:
                ndnfree = ndnmaxsize - ndncursize
                ndnfreeratio = "{:.1f}".format(ndnfree / ndnmaxsize * 100)

            result['ndncache'] = {
                'hit_ratio': ndnratio,
                'free': convert_bytes(str(ndnfree)),
                'free_percentage': ndnfreeratio,
                'count': ndncount,
                'evictions': ndnevictions
            }
        # In case, the user enabled NDN cache but still have not restarted the instance
        except IndexError:
            ndn_cache_enabled = False

    # Build the backend results
    for be in backend_objs:
        be_name = be.rdn
        be_suffix = be.get_suffix()
        monitor = be.get_monitor()
        all_attrs = monitor.get_status()

        # Process entry cache stats
        entcur = int(all_attrs['currententrycachesize'][0])
        entmax = int(all_attrs['maxentrycachesize'][0])
        entcnt = int(all_attrs['currententrycachecount'][0])
        entratio = all_attrs['entrycachehitratio'][0]
        entfree = entmax - entcur
        entfreep = "{:.1f}".format(entfree / entmax * 100)
        if entcnt == 0:
            entsize = 0
        else:
            entsize = int(entcur / entcnt)

        if ldbm_monitor.inst_db_impl == DB_IMPL_BDB:
            # Process DN cache stats
            dncur = int(all_attrs['currentdncachesize'][0])
            dnmax = int(all_attrs['maxdncachesize'][0])
            dncnt = int(all_attrs['currentdncachecount'][0])
            dnratio = all_attrs['dncachehitratio'][0]
            dnfree = dnmax - dncur
            dnfreep = "{:.1f}".format(dnfree / dnmax * 100)
            if dncnt == 0:
                dnsize = 0
            else:
                dnsize = int(dncur / dncnt)

        # Build the backend result
        result['backends'][be_name] = {
            'suffix': be_suffix,
            'entry_cache_count': all_attrs['currententrycachecount'][0],
            'entry_cache_free': convert_bytes(str(entfree)),
            'entry_cache_free_percentage': entfreep,
            'entry_cache_size': convert_bytes(str(entsize)),
            'entry_cache_hit_ratio': entratio,
            'indexes': []
        }
        if ldbm_monitor.inst_db_impl == DB_IMPL_BDB:
            backend = result['backends'][be_name]
            backend['dn_cache_count'] = all_attrs['currentdncachecount'][0]
            backend['dn_cache_free'] = convert_bytes(str(dnfree))
            backend['dn_cache_free_percentage'] = dnfreep
            backend['dn_cache_size'] = convert_bytes(str(dnsize))
            backend['dn_cache_hit_ratio'] = dnratio

        # Process indexes if requested
        if args.indexes:
            index = {}
            index_name = ''
            for attr, val in all_attrs.items():
                if attr.startswith('dbfile'):
                    if attr.startswith("dbfilename-"):
                        if index_name != '':
                            # Update backend index list
                            result['backends'][be_name]['indexes'].append(index)
                        index_name = val[0].split('/')[1]
                        index = {'name': index_name}
                    elif attr.startswith('dbfilecachehit-'):
                        index['cachehit'] = val[0]
                    elif attr.startswith('dbfilecachemiss-'):
                        index['cachemiss'] = val[0]
                    elif attr.startswith('dbfilepagein-'):
                        index['pagein'] = val[0]
                    elif attr.startswith('dbfilepageout-'):
                        index['pageout'] = val[0]
            if index_name != '':
                # Update backend index list
                result['backends'][be_name]['indexes'].append(index)

    # Return the report
    if args.json:
        log.info(json.dumps(result, indent=4))
    else:
        log.info("DB Monitor Report: " + result['date'])
        log.info("--------------------------------------------------------")
        if ldbm_monitor.inst_db_impl == DB_IMPL_BDB:
            log.info("Database Cache:")
            log.info(" - Cache Hit Ratio:     {}%".format(result['dbcache']['hit_ratio']))
            log.info(" - Free Space:          {}".format(result['dbcache']['free']))
            log.info(" - Free Percentage:     {}%".format(result['dbcache']['free_percentage']))
            log.info(" - RO Page Drops:       {}".format(result['dbcache']['roevicts']))
            log.info(" - Pages In:            {}".format(result['dbcache']['pagein']))
            log.info(" - Pages Out:           {}".format(result['dbcache']['pageout']))
            log.info("")

        if ndn_cache_enabled:
            log.info("Normalized DN Cache:")
            log.info(" - Cache Hit Ratio:     {}%".format(result['ndncache']['hit_ratio']))
            log.info(" - Free Space:          {}".format(result['ndncache']['free']))
            log.info(" - Free Percentage:     {}%".format(result['ndncache']['free_percentage']))
            log.info(" - DN Count:            {}".format(result['ndncache']['count']))
            log.info(" - Evictions:           {}".format(result['ndncache']['evictions']))
            log.info("")

        log.info("Backends:")
        for be_name, attr_dict in result['backends'].items():
            log.info(f"  - {attr_dict['suffix']} ({be_name}):")
            log.info("    - Entry Cache Hit Ratio:        {}%".format(attr_dict['entry_cache_hit_ratio']))
            log.info("    - Entry Cache Count:            {}".format(attr_dict['entry_cache_count']))
            log.info("    - Entry Cache Free Space:       {}".format(attr_dict['entry_cache_free']))
            log.info("    - Entry Cache Free Percentage:  {}%".format(attr_dict['entry_cache_free_percentage']))
            log.info("    - Entry Cache Average Size:     {}".format(attr_dict['entry_cache_size']))
            if ldbm_monitor.inst_db_impl == DB_IMPL_BDB:
                log.info("    - DN Cache Hit Ratio:           {}%".format(attr_dict['dn_cache_hit_ratio']))
                log.info("    - DN Cache Count:               {}".format(attr_dict['dn_cache_count']))
                log.info("    - DN Cache Free Space:          {}".format(attr_dict['dn_cache_free']))
                log.info("    - DN Cache Free Percentage:     {}%".format(attr_dict['dn_cache_free_percentage']))
                log.info("    - DN Cache Average Size:        {}".format(attr_dict['dn_cache_size']))
            if len(result['backends'][be_name]['indexes']) > 0:
                log.info("    - Indexes:")
                for index in result['backends'][be_name]['indexes']:
                    log.info("      - Index:      {}".format(index['name']))
                    log.info("      - Cache Hit:  {}".format(index['cachehit']))
                    log.info("      - Cache Miss: {}".format(index['cachemiss']))
                    log.info("      - Page In:    {}".format(index['pagein']))
                    log.info("      - Page Out:   {}".format(index['pageout']))
                    log.info("")
            log.info("")


def memberof_monitor(inst, basedn, log, args):
    memberof_monitor = MonitorMemberOf(inst)
    _format_status(log, memberof_monitor, args.json)


def create_parser(subparsers):
    monitor_parser = subparsers.add_parser('monitor', help="Monitor the state of the instance", formatter_class=CustomHelpFormatter)
    subcommands = monitor_parser.add_subparsers(help='action')

    server_parser = subcommands.add_parser('server', help="Displays the server statistics, connections, and operations", formatter_class=CustomHelpFormatter)
    server_parser.set_defaults(func=monitor)
    server_parser.add_argument('-r', '--just-resources', action='store_true',
                               default=False,
                               help="Just display the server resources being consumed")

    dbmon_parser = subcommands.add_parser('dbmon', help="Monitor all database statistics in a single report", formatter_class=CustomHelpFormatter)
    dbmon_parser.set_defaults(func=db_monitor)
    dbmon_parser.add_argument('-b', '--backends', help="Specifies a list of space-separated backends to monitor. Default is all backends.")
    dbmon_parser.add_argument('-x', '--indexes', action='store_true', default=False, help="Shows index stats for each backend")

    ldbm_parser = subcommands.add_parser('ldbm', help="Monitor the LDBM statistics, such as dbcache", formatter_class=CustomHelpFormatter)
    ldbm_parser.set_defaults(func=ldbm_monitor)

    backend_parser = subcommands.add_parser('backend', help="Monitor the behavior of a backend database", formatter_class=CustomHelpFormatter)
    backend_parser.add_argument('backend', nargs='?', help="The optional name of the backend to monitor")
    backend_parser.set_defaults(func=backend_monitor)

    snmp_parser = subcommands.add_parser('snmp', help="Displays the SNMP statistics", formatter_class=CustomHelpFormatter)
    snmp_parser.set_defaults(func=snmp_monitor)

    chaining_parser = subcommands.add_parser('chaining', help="Monitor database chaining statistics", formatter_class=CustomHelpFormatter)
    chaining_parser.add_argument('backend', nargs='?', help="The optional name of the chaining backend to monitor")
    chaining_parser.set_defaults(func=chaining_monitor)

    disk_parser = subcommands.add_parser('disk', help="Displays the disk space statistics. All values are in bytes.", formatter_class=CustomHelpFormatter)
    disk_parser.set_defaults(func=disk_monitor)

    deferred_mo_parser = subcommands.add_parser('memberof', help="Displays deferred memberof plugin stats.", formatter_class=CustomHelpFormatter)
    deferred_mo_parser.set_defaults(func=memberof_monitor)


# --- BEGIN COPYRIGHT BLOCK ---
# Copyright (C) 2015 Red Hat, Inc.
# All rights reserved.
#
# License: GPL (version 3 or any later version).
# See LICENSE for details.
# --- END COPYRIGHT BLOCK ---
#
from lib389._constants import *
from lib389 import DirSrv, Entry
import pytest
import logging

logging.getLogger(__name__).setLevel(logging.DEBUG)
log = logging.getLogger(__name__)

INSTANCE_PORT = 54321
INSTANCE_SERVERID = 'standalone'
TEST_USER = 'uid=test,%s' % DEFAULT_SUFFIX
TEST_GROUP = 'cn=testgroup,%s' % DEFAULT_SUFFIX


class TopologyInstance(object):
    def __init__(self, instance):
        instance.open()
        self.instance = instance


@pytest.fixture(scope="module")
def topology(request):
    instance = DirSrv(verbose=False)
    instance.log.debug("Instance allocated")
    args = {SER_HOST: LOCALHOST,
            SER_PORT: INSTANCE_PORT,
            SER_SERVERID_PROP: INSTANCE_SERVERID}
    instance.allocate(args)
    if instance.exists():
        instance.delete()
    instance.create()
    instance.open()

    def fin():
        if instance.exists():
            instance.delete()
    request.addfinalizer(fin)

    return TopologyInstance(instance)


@pytest.fixture(scope="module")
def add_user(topology):
    """Create a user entry"""

    log.info('Create a user entry: %s' % TEST_USER)
    uentry = Entry(TEST_USER)
    uentry.setValues('objectclass', 'top', 'extensibleobject')
    uentry.setValues('uid', 'test')
    topology.instance.add_s(uentry)
    # topology.instance.log.debug("Created user entry as:" ,uentry.dn)


@pytest.fixture(scope="module")
def add_group(topology):
    """Create a group for the user to have some rights to"""

    log.info('Create a group entry: %s' % TEST_GROUP)
    gentry = Entry(TEST_GROUP)
    gentry.setValues('objectclass', 'top', 'extensibleobject')
    gentry.setValues('cn', 'testgroup')
    topology.instance.add_s(gentry)


def test_effective_rights(topology, add_user, add_group):
    """Run an effective rights search
    and compare actual results with expected
    """

    log.info('Search for effective rights with get_effective_rights() '
             'function')
    result = topology.instance.get_effective_rights(TEST_USER,
                                                    filterstr='(cn=testgroup)',
                                                    attrlist=['cn'])

    rights = result[0]
    log.info('Assert that "attributeLevelRights: cn:rsc" and '
             '"entryLevelRights: v"')
    assert rights.getValue('attributeLevelRights') == 'cn:rsc'
    assert rights.getValue('entryLevelRights') == 'v'


if __name__ == "__main__":
    CURRENT_FILE = os.path.realpath(__file__)
    pytest.main("-s -v %s" % CURRENT_FILE)



# --- BEGIN COPYRIGHT BLOCK ---
# Copyright (C) 2016 Red Hat, Inc.
# All rights reserved.
#
# License: GPL (version 3 or any later version).
# See LICENSE for details.
# --- END COPYRIGHT BLOCK ---

import pytest
import sys
import io
from lib389.cli_base import LogCapture
from lib389._constants import *
from lib389.tests.topologies import create_topology, DEBUGGING
from lib389.configurations import get_sample_entries


@pytest.fixture(scope="module")
def topology(request):
    topology = create_topology({ReplicaRole.STANDALONE: 1}, None)

    def fin():
        if DEBUGGING:
            topology.standalone.stop()
        else:
            topology.standalone.delete()
    request.addfinalizer(fin)

    topology.logcap = LogCapture()
    return topology


@pytest.fixture(scope="module")
def topology_be_latest(request):
    topology = create_topology({ReplicaRole.STANDALONE: 1}, None)
    topology.standalone.backends.create(properties={
        'cn': 'userRoot',
        'nsslapd-suffix': DEFAULT_SUFFIX,
    })
    # Now apply sample entries
    centries = get_sample_entries(INSTALL_LATEST_CONFIG)
    cent = centries(topology.standalone, DEFAULT_SUFFIX)
    cent.apply()

    def fin():
        if DEBUGGING:
            topology.standalone.stop()
        else:
            topology.standalone.delete()
    request.addfinalizer(fin)

    topology.logcap = LogCapture()
    return topology


@pytest.fixture(scope="module")
def topology_be_001003006(request):
    topology = create_topology({ReplicaRole.STANDALONE: 1}, None)
    topology.standalone.backends.create(properties={
        'cn': 'userRoot',
        'nsslapd-suffix': DEFAULT_SUFFIX,
    })
    # Now apply sample entries
    centries = get_sample_entries('001003006')
    cent = centries(topology.standalone, DEFAULT_SUFFIX)
    cent.apply()

    def fin():
        if DEBUGGING:
            topology.standalone.stop()
        else:
            topology.standalone.delete()
    request.addfinalizer(fin)

    topology.logcap = LogCapture()
    return topology


def check_output(some_string, missing=False, ignorecase=True):
    """Check the output of captured STDOUT.  This assumes "sys.stdout = io.StringIO()"
    otherwise there would be nothing to read.  Flush IO after performing check.
    :param some_string - text, or list of strings, to search for in output
    :param missing - test if some_string is NOT present in output
    :param ignorecase - Set whether to ignore the character case in both the output
                        and some_string
    """
    output = sys.stdout.getvalue()
    is_list = isinstance(some_string, list)

    if ignorecase:
        output = output.lower()
        if is_list:
            some_string = [text.lower() for text in some_string]
        else:
            some_string = some_string.lower()

    if missing:
        if is_list:
            for text in some_string:
                assert(text not in output)
        else:
            assert(some_string not in output)
    else:
        if is_list:
            for text in some_string:
                assert(text in output)
        else:
            assert(some_string in output)

    # Clear the buffer
    sys.stdout = io.StringIO()



# --- BEGIN COPYRIGHT BLOCK ---
# Copyright (C) 2025 Red Hat, Inc.
# All rights reserved.
#
# License: GPL (version 3 or any later version).
# See LICENSE for details.
# --- END COPYRIGHT BLOCK ---

import ldap
from ldap import filter as ldap_filter
from lib389._mapped_object import DSLdapObject, DSLdapObjects, _normalise_attrs
from lib389.backend import Backends
from lib389.config import Config
from lib389.idm.account import Account
from lib389.idm.nscontainer import nsContainers, nsContainer
from lib389.idm.user import (
    BasicUserAccounts,
    nsUserAccounts,
    TraditionalUserAccounts,
)
from lib389.idm.services import ServiceAccounts
from lib389.cos import CosPointerDefinitions, CosPointerDefinition, CosTemplates


POLICY_TYPE_GLOBAL = 'Global Policy'
POLICY_TYPE_USER = 'User Policy'
POLICY_TYPE_SUBTREE = 'Subtree Policy'

SYNTAX_POLICY_ATTRS = [
    'passwordchecksyntax',
    'passwordminlength',
    'passwordmindigits',
    'passwordminalphas',
    'passwordminuppers',
    'passwordminlowers',
    'passwordminspecials',
    'passwordmin8bit',
    'passwordmaxrepeats',
    'passwordpalindrome',
    'passwordmaxsequence',
    'passwordmaxseqsets',
    'passwordmaxclasschars',
    'passwordmincategories',
    'passwordmintokenlength',
    'passwordbadwords',
    'passworduserattributes',
    'passworddictcheck',
    'passworddictpath',
]

GLOBAL_ONLY_POLICY_ATTRS = [
    'passwordisglobalpolicy',
    'nsslapd-pwpolicy-local',
    'nsslapd-allow-hashed-passwords',
    'nsslapd-pwpolicy-inherit-global',
]

USER_TYPE_ACCOUNTS = {
    'posix': nsUserAccounts,
    'basic': BasicUserAccounts,
    'service': ServiceAccounts,
    'traditional': TraditionalUserAccounts,
}

# Defaults from pwpolicy_init_defaults() in ldap/servers/slapd/libglobs.c
LOCAL_PW_POLICY_DEFAULTS = {
    'passwordchange': 'on',
    'passwordmustchange': 'off',
    'passwordchecksyntax': 'off',
    'passwordexp': 'off',
    'passwordsendexpiringtime': 'off',
    'passwordminlength': '8',
    'passwordmindigits': '0',
    'passwordminalphas': '0',
    'passwordminuppers': '0',
    'passwordminlowers': '0',
    'passwordminspecials': '0',
    'passwordmin8bit': '0',
    'passwordmaxrepeats': '0',
    'passwordmincategories': '3',
    'passwordmintokenlength': '3',
    'passwordmaxage': '8640000',
    'passwordminage': '0',
    'passwordwarning': '86400',
    'passwordhistory': 'off',
    'passwordinhistory': '6',
    'passwordlockout': 'off',
    'passwordmaxfailure': '3',
    'passwordunlock': 'on',
    'passwordlockoutduration': '3600',
    'passwordresetfailurecount': '600',
    'passwordTPRMaxUse': '-1',
    'passwordTPRDelayExpireAt': '-1',
    'passwordTPRDelayValidFrom': '-1',
    'passwordgracelimit': '0',
    'passwordadmindn': '',
    'passwordadminskipinfoupdate': 'off',
    'passwordtrackupdatetime': 'off',
    'passwordpalindrome': 'off',
    'passwordmaxsequence': '0',
    'passwordmaxseqsets': '0',
    'passwordmaxclasschars': '0',
    'passwordbadwords': '',
    'passworduserattributes': '',
    'passworddictcheck': 'off',
    'passworddictpath': '',
    'passwordstoragescheme': '',
}


def _decode_cn_dn_fragment(fragment):
    """Decode LDAP escaped characters in a cn-stored DN fragment."""
    return (fragment.replace('\\3D', '=').replace('\\3d', '=')
            .replace('\\2C', ',').replace('\\2c', ','))


def _parse_policy_target_from_cn(cn_value):
    """Return the policy target DN encoded in a policy entry cn value."""
    if not cn_value:
        return None
    cn_value = cn_value.strip("'").strip('"')
    cn_lower = cn_value.lower()
    prefixes = (
        'cn=nspwpolicyentry_user,',
        'cn=nspwpolicyentry_subtree,',
    )
    if not any(cn_lower.startswith(p) for p in prefixes):
        return None
    try:
        dn_comps = ldap.dn.explode_dn(cn_value)
        dn_comps.pop(0)
        if dn_comps:
            return ','.join(dn_comps)
    except ldap.DECODING_ERROR:
        # Decoding errors are ok, just proceed
        pass

    remainder = _decode_cn_dn_fragment(cn_value[cn_value.index(',') + 1:])
    try:
        return ldap.dn.dn2str(ldap.dn.str2dn(remainder))
    except ldap.DECODING_ERROR:
        # It's ok, just return the remainder
        return remainder


def _classify_policy_from_pwdpolicysubentry(pwp_dn, pwp_entry):
    """Return policy type, source DN, and target DN for a local policy entry."""
    cn = pwp_entry.get_attr_val_utf8('cn') or ''
    cn_l = cn.lower()
    if 'nspwpolicyentry_user' in cn_l:
        policy_type = POLICY_TYPE_USER
    elif 'nspwpolicyentry_subtree' in cn_l:
        policy_type = POLICY_TYPE_SUBTREE
    else:
        raise ValueError(
            'Unable to determine password policy type for entry "{}"'.format(pwp_dn))
    target = _parse_policy_target_from_cn(cn)
    if not target:
        raise ValueError(
            'Unable to parse password policy target from policy entry "{}"'.format(pwp_dn))
    return policy_type, pwp_dn, target


def _attr_is_on(value):
    if value is None:
        return False
    return str(value).lower() in ('on', 'true', '1', 'yes')


def _should_show_inherited_value(value):
    """Return True when an inherited setting should appear in local policy output."""
    if value is None:
        return False
    text = str(value).strip()
    if not text or text == '0':
        return False
    if text.lower() == 'off':
        return False
    return True


def _first_attr_value(attrs, attr_name, default=''):
    if not attrs:
        return default
    values = attrs.get(attr_name, [])
    if values:
        return values[0]
    attr_lower = attr_name.lower()
    for key, vals in attrs.items():
        if key.lower() == attr_lower and vals:
            return vals[0]
    return default


def _make_attr_entry(value, inherited=False):
    return {'values': ['' if value is None else str(value)], 'inherited': inherited}


class PwPolicyManager(object):
    """Manages user, subtree and global password policies

    :param instance: An instance
    :type instance: lib389.DirSrv
    """

    def __init__(self, instance):
        self._instance = instance
        self.log = instance.log
        self.arg_to_attr = {
            'pwdlocal': 'nsslapd-pwpolicy-local',
            'pwdscheme': 'passwordstoragescheme',
            'pwdchange': 'passwordchange',
            'pwdmustchange': 'passwordmustchange',
            'pwdhistory': 'passwordhistory',
            'pwdhistorycount': 'passwordinhistory',
            'pwdadmin': 'passwordadmindn',
            'pwdadminskipupdates': 'passwordadminskipinfoupdate',
            'pwdtrack': 'passwordtrackupdatetime',
            'pwdwarning': 'passwordwarning',
            'pwdisglobal': 'passwordisglobalpolicy',
            'pwdexpire': 'passwordexp',
            'pwdmaxage': 'passwordmaxage',
            'pwdminage': 'passwordminage',
            'pwdgracelimit': 'passwordgracelimit',
            'pwdsendexpiring': 'passwordsendexpiringtime',
            'pwdlockout': 'passwordlockout',
            'pwdunlock': 'passwordunlock',
            'pwdlockoutduration': 'passwordlockoutduration',
            'pwdmaxfailures': 'passwordmaxfailure',
            'pwdresetfailcount': 'passwordresetfailurecount',
            'pwdchecksyntax': 'passwordchecksyntax',
            'pwdminlen': 'passwordminlength',
            'pwdmindigits': 'passwordmindigits',
            'pwdminalphas': 'passwordminalphas',
            'pwdminuppers': 'passwordminuppers',
            'pwdminlowers': 'passwordminlowers',
            'pwdminspecials': 'passwordminspecials',
            'pwdmin8bits': 'passwordmin8bit',
            'pwdmaxrepeats': 'passwordmaxrepeats',
            'pwdpalindrome': 'passwordpalindrome',
            'pwdmaxseq': 'passwordmaxsequence',
            'pwdmaxseqsets': 'passwordmaxseqsets',
            'pwdmaxclasschars': 'passwordmaxclasschars',
            'pwdmincatagories': 'passwordmincategories',
            'pwdmintokenlen': 'passwordmintokenlength',
            'pwdbadwords': 'passwordbadwords',
            'pwduserattrs': 'passworduserattributes',
            'pwddictcheck': 'passworddictcheck',
            'pwddictpath': 'passworddictpath',
            'pwdallowhash': 'nsslapd-allow-hashed-passwords',
            'pwpinheritglobal': 'nsslapd-pwpolicy-inherit-global',
            'pwptprmaxuse': 'passwordTPRMaxUse',
            'pwptprdelayexpireat': 'passwordTPRDelayExpireAt',
            'pwptprdelayvalidfrom': 'passwordTPRDelayValidFrom'
        }

    def policy_attribute_names(self, include_global_only=False):
        """Return sorted password policy attribute names for CLI output."""
        names = sorted(set(self.arg_to_attr.values()))
        if include_global_only:
            for attr in GLOBAL_ONLY_POLICY_ATTRS:
                if attr not in names:
                    names.append(attr)
            names.sort()
        return names

    def _local_policy_attribute_names(self):
        """Return password policy attribute names stored on local policy entries."""
        return sorted(set(self.arg_to_attr.values()) | {'passwordstoragescheme'})

    def _report_attribute_names(self, report):
        """Return attribute names to emit for a policy report."""
        if report['policy_type'] == POLICY_TYPE_GLOBAL:
            return self.policy_attribute_names(include_global_only=True)
        local = []
        inherited = []
        for attr in sorted(report['attrs'].keys()):
            if report['attrs'][attr].get('inherited'):
                inherited.append(attr)
            else:
                local.append(attr)
        return local + inherited

    def resolve_user_dn(self, basedn, user_type, selector, parent_basedn=None):
        """Resolve a dsidm user selector to an entry DN."""
        if not selector:
            raise ValueError('A user identifier is required')
        accounts_class = USER_TYPE_ACCOUNTS.get(user_type or 'posix', nsUserAccounts)
        if parent_basedn is not None:
            accounts = accounts_class(self._instance, parent_basedn, rdn=None)
        else:
            accounts = accounts_class(self._instance, basedn)
        if not accounts.exists(selector):
            raise ValueError('User "{}" was not found'.format(selector))
        return accounts.get(selector).dn

    def get_effective_policy(self, basedn, user_type, selector, parent_basedn=None):
        """Resolve the effective password policy for a user account."""
        user_dn = self.resolve_user_dn(basedn, user_type, selector, parent_basedn)
        user_entry = Account(self._instance, user_dn)
        user_attrs = user_entry.get_all_attrs_utf8()
        if not user_attrs:
            raise ValueError('User entry "{}" was not found'.format(user_dn))

        if self._instance.config.get_attr_val_utf8_l('nsslapd-pwpolicy-local') == 'on':
            pwp_subentry_vals = user_attrs.get('pwdpolicysubentry', [])
            pwp_dn = pwp_subentry_vals[0] if pwp_subentry_vals else None
        else:
            pwp_dn = None

        global_attrs_raw = self._get_global_policy_attrs()

        if pwp_dn is None:
            return {
                'dn': user_dn,
                'policy_type': POLICY_TYPE_GLOBAL,
                'policy_dn': Config(self._instance).dn,
                'policy_target': 'global',
                'attrs': self._build_global_effective_attrs(global_attrs_raw),
            }

        pwp_entry = PwPolicyEntry(self._instance, pwp_dn)
        if not pwp_entry.exists():
            raise ValueError(
                'Assigned password policy entry "{}" could not be found'.format(pwp_dn))
        if not pwp_entry.present('objectClass', 'passwordpolicy'):
            raise ValueError(
                'Assigned password policy entry "{}" is not a valid password policy'.format(pwp_dn))

        policy_type, policy_dn, policy_target = _classify_policy_from_pwdpolicysubentry(
            pwp_dn, pwp_entry)
        policy_attrs_raw = _normalise_attrs(pwp_entry.get_all_attrs_utf8())
        effective = self._build_local_effective_attrs(policy_attrs_raw, global_attrs_raw)

        return {
            'dn': user_dn,
            'policy_type': policy_type,
            'policy_dn': policy_dn,
            'policy_target': policy_target,
            'attrs': effective,
        }

    def _get_global_policy_attrs(self):
        config = Config(self._instance)
        attr_names = self.policy_attribute_names(include_global_only=True)
        attrs = _normalise_attrs(config.get_attrs_vals_utf8(attr_names))
        # Some config attributes may be returned only under schema canonical names.
        all_attrs = _normalise_attrs(config.get_all_attrs_utf8())
        for name in attr_names:
            if name not in attrs or not attrs[name]:
                if name in all_attrs and all_attrs[name]:
                    attrs[name] = all_attrs[name]
        return attrs

    def _build_global_effective_attrs(self, global_attrs_raw):
        effective = {}
        for attr in self.policy_attribute_names(include_global_only=True):
            value = _first_attr_value(global_attrs_raw, attr, LOCAL_PW_POLICY_DEFAULTS.get(attr, ''))
            effective[attr] = _make_attr_entry(value, inherited=False)
        return effective

    def _build_local_effective_attrs(self, policy_attrs_raw, global_attrs_raw):
        """Build attrs for a local policy: explicit entry values plus inherited syntax."""
        policy_attrs_raw = _normalise_attrs(policy_attrs_raw)
        global_attrs_raw = _normalise_attrs(global_attrs_raw)
        effective = {}
        local_attrs = set()

        for attr in self._local_policy_attribute_names():
            if attr in policy_attrs_raw and policy_attrs_raw[attr]:
                effective[attr] = _make_attr_entry(
                    _first_attr_value(policy_attrs_raw, attr), inherited=False)
                local_attrs.add(attr)

        if 'passwordchecksyntax' in local_attrs:
            local_syntax_on = _attr_is_on(effective['passwordchecksyntax']['values'][0])
        else:
            local_syntax_on = False

        inherit_global = _attr_is_on(
            _first_attr_value(global_attrs_raw, 'nsslapd-pwpolicy-inherit-global', 'off'))
        global_syntax_on = _attr_is_on(
            _first_attr_value(global_attrs_raw, 'passwordchecksyntax', 'off'))

        if not local_syntax_on and inherit_global and global_syntax_on:
            for attr in SYNTAX_POLICY_ATTRS:
                if attr == 'passwordchecksyntax' or attr in local_attrs:
                    continue
                value = _first_attr_value(global_attrs_raw, attr, '')
                if not _should_show_inherited_value(value):
                    continue
                effective[attr] = _make_attr_entry(value, inherited=True)

        return effective

    def format_report_text(self, report):
        """Format an effective policy report for human-readable CLI output."""
        lines = [
            'dn: {}'.format(report['dn']),
            'passwordPolicy: {}'.format(report['policy_type']),
            'passwordPolicyDN: {}'.format(report['policy_dn']),
            'passwordPolicyTarget: {}'.format(report['policy_target']),
            '------------------- Policy Settings -------------------',
        ]
        for attr in self._report_attribute_names(report):
            item = report['attrs'][attr]
            value = item['values'][0] if item['values'] else ''
            suffix = ' (inherited)' if item.get('inherited') else ''
            lines.append('{}: {}{}'.format(attr, value, suffix))
        return '\n'.join(lines) + '\n'

    def format_report_json(self, report):
        """Format an effective policy report as a JSON-serializable dict."""
        attrs_out = {}
        for attr in self._report_attribute_names(report):
            item = report['attrs'][attr]
            values = list(item['values'])
            if item.get('inherited'):
                values.append('inherited')
            attrs_out[attr] = values
        return {
            'dn': report['dn'],
            'policy_type': report['policy_type'],
            'policy_dn': report['policy_dn'],
            'policy_target': report['policy_target'],
            'attrs': attrs_out,
        }

    def is_subtree_policy(self, dn):
        """Check if a subtree password policy exists for a given entry DN.

        A subtree policy is indicated by the presence of any CoS template
        (under `cn=nsPwPolicyContainer,<dn>`) that has a `pwdpolicysubentry`
        attribute pointing to an existing entry with objectClass `passwordpolicy`.

        :param dn: Entry DN to check for subtree policy
        :type dn: str

        :returns: True if a subtree policy exists, False otherwise
        :rtype: bool
        """
        try:
            container_basedn = 'cn=nsPwPolicyContainer,{}'.format(dn)
            templates = CosTemplates(self._instance, container_basedn).list()
            for tmpl in templates:
                pwp_dn = tmpl.get_attr_val_utf8('pwdpolicysubentry')
                if not pwp_dn:
                    continue
                # Validate that the referenced entry exists and is a passwordpolicy
                pwp_entry = PwPolicyEntry(self._instance, pwp_dn)
                if pwp_entry.exists() and pwp_entry.present('objectClass', 'passwordpolicy'):
                    return True
        except ldap.LDAPError:
            pass
        return False

    def is_user_policy(self, dn):
        """Check if the entry has a user password policy.

        A user policy is indicated by the target entry having a
        `pwdpolicysubentry` attribute that points to an existing
        entry with objectClass `passwordpolicy`.

        :param dn: Entry DN to check
        :type dn: str

        :returns: True if the entry has a user policy, False otherwise
        :rtype: bool
        """
        try:
            entry = Account(self._instance, dn)
            if not entry.exists():
                return False
            pwp_dn = entry.get_attr_val_utf8('pwdpolicysubentry')
            if not pwp_dn:
                return False
            pwp_entry = PwPolicyEntry(self._instance, pwp_dn)
            return pwp_entry.exists() and pwp_entry.present('objectClass', 'passwordpolicy')
        except ldap.LDAPError:
            return False

    def create_user_policy(self, dn, properties):
        """Creates all entries which are needed for the user
        password policy

        :param dn: Entry DN for the subtree pwpolicy
        :type dn: str
        :param properties: A dict with password policy settings
        :type properties: dict

        :returns: PwPolicyEntry instance
        """

        # Verify target dn exists before getting started
        user_entry = Account(self._instance, dn)
        if not user_entry.exists():
            raise ValueError('Can not create user password policy because the target dn does not exist')

        dn_comps = ldap.dn.explode_dn(user_entry.dn)
        dn_comps.pop(0)
        parentdn = ",".join(dn_comps)

        # Create the pwp container if needed
        pwp_containers = nsContainers(self._instance, basedn=parentdn)
        pwp_container = pwp_containers.ensure_state(properties={'cn': 'nsPwPolicyContainer'}, strict=True)

        # Create or update the policy entry
        properties['cn'] = 'cn=nsPwPolicyEntry_user,%s' % dn
        pwp_entries = PwPolicyEntries(self._instance, pwp_container.dn)
        pwp_entry = pwp_entries.ensure_state(properties=properties, strict=True)
        try:
            # Add policy to the entry if needed
            user_pwp_dn = user_entry.get_attr_val_utf8('pwdpolicysubentry')
            if user_pwp_dn != pwp_entry.dn:
                user_entry.replace('pwdpolicysubentry', pwp_entry.dn)
        except ldap.LDAPError as e:
            # failure, undo what we have done
            pwp_entry.delete()
            raise e

        # make sure that local policies are enabled
        self.set_global_policy({'nsslapd-pwpolicy-local': 'on'})

        return pwp_entry

    def create_subtree_policy(self, dn, properties):
        """Creates all entries which are needed for the subtree
        password policy

        :param dn: Entry DN for the subtree pwpolicy
        :type dn: str
        :param properties: A dict with password policy settings
        :type properties: dict

        :returns: PwPolicyEntry instance
        """

        # Verify target dn exists before getting started
        subtree_entry = Account(self._instance, dn)
        if not subtree_entry.exists():
            raise ValueError('Can not create subtree password policy because the target dn does not exist')

        # Create the pwp container if needed
        pwp_containers = nsContainers(self._instance, basedn=dn)
        pwp_container = pwp_containers.ensure_state(properties={'cn': 'nsPwPolicyContainer'}, strict=True)

        # Create or update the policy entry
        properties['cn'] = 'cn=nsPwPolicyEntry_subtree,%s' % dn
        pwp_entries = PwPolicyEntries(self._instance, pwp_container.dn)
        pwp_entry = pwp_entries.ensure_state(properties=properties, strict=True)

        # Ensure the CoS template entry (nsPwTemplateEntry) that points to the
        # password policy entry
        cos_templates = CosTemplates(self._instance, pwp_container.dn)
        cos_template = cos_templates.ensure_state(properties={
            'cosPriority': '1',
            'pwdpolicysubentry': pwp_entry.dn,
            'cn': 'cn=nsPwTemplateEntry,%s' % dn
        }, strict=True)

        # Ensure the CoS specification entry at the subtree level
        cos_pointer_defs = CosPointerDefinitions(self._instance, dn)
        cos_pointer_def = cos_pointer_defs.ensure_state(properties={
            'cosAttribute': 'pwdpolicysubentry default operational-default',
            'cosTemplateDn': cos_template.dn,
            'cn': 'nsPwPolicy_CoS'
        }, strict=True)

        # make sure that local policies are enabled
        self.set_global_policy({'nsslapd-pwpolicy-local': 'on'})

        # Capture the outcome of each ensure_state call
        ensure_status = [
            pwp_container.ensure_status,
            pwp_entry.ensure_status,
            cos_template.ensure_status,
            cos_pointer_def.ensure_status,
        ]

        # Determine overall status
        if all(s == DSLdapObject.ENSURE_UNCHANGED for s in ensure_status):
            pwp_entry._ensure_status = DSLdapObject.ENSURE_UNCHANGED
        elif all(s == DSLdapObject.ENSURE_ADDED for s in ensure_status):
            pwp_entry._ensure_status = DSLdapObject.ENSURE_ADDED
        else:
            # If any component was added/updated, consider it UPDATED (repair or modification)
            pwp_entry._ensure_status = DSLdapObject.ENSURE_UPDATED

        return pwp_entry

    def get_pwpolicy_entry(self, dn):
        """Get a local password policy entry

        :param dn: Entry DN for the local pwpolicy
        :type dn: str

        :returns: PwPolicyEntry instance
        """

        entry = Account(self._instance, dn)
        if not entry.exists():
            raise ValueError('Can not get the password policy entry because the target dn does not exist')

        try:
            Backends(self._instance).get(dn)
            # The DN is a base suffix, it has no parent
            parentdn = dn
        except:
            # Ok, this is a not a suffix, get the parent DN
            dn_comps = ldap.dn.explode_dn(entry.dn)
            dn_comps.pop(0)
            parentdn = ",".join(dn_comps)

        # Get the parent's policies
        pwp_entries = PwPolicyEntries(self._instance, parentdn)
        policies = pwp_entries.list()

        for policy in policies:
            # Sometimes, the cn value includes quotes (for example, after migration from pre-CLI version).
            # We need to strip them as python-ldap doesn't expect them
            dn_comps_str = policy.get_attr_val_utf8_l('cn').strip("\'").strip("\"")
            try:
                dn_comps = ldap.dn.explode_dn(dn_comps_str)
                dn_comps.pop(0)
                pwp_dn = ",".join(dn_comps)
                if pwp_dn == dn.lower():
                    # This DN does have a policy associated with it
                    return policy
            except ldap.DECODING_ERROR:
                # This is some kind of custom password policy
                continue

        # Did not find a policy for this entry
        raise ValueError("No password policy was found for this entry")

    def delete_local_policy(self, dn):
        """Delete a local password policy entry

        :param dn: Entry DN for the local pwpolicy
        :type dn: str
        """

        subtree = False

        # Verify target dn exists, and has a policy
        entry = Account(self._instance, dn)
        if not entry.exists():
            raise ValueError('The target entry dn does not exist')
        pwp_entry = self.get_pwpolicy_entry(entry.dn)

        # Subtree or user policy?
        if self.is_subtree_policy(entry.dn):
            parentdn = dn
            subtree = True
        elif self.is_user_policy(entry.dn):
            dn_comps = ldap.dn.explode_dn(dn)
            dn_comps.pop(0)
            parentdn = ",".join(dn_comps)
        else:
            raise ValueError('The target entry dn does not have a password policy')

        # Starting deleting the policy, ignore the parts that might already have been removed
        pwp_container = nsContainer(self._instance, 'cn=nsPwPolicyContainer,%s' % parentdn)
        if subtree:
            try:
                # Delete template
                cos_templates = CosTemplates(self._instance, pwp_container.dn)
                cos_template = cos_templates.get('cn=nsPwTemplateEntry,%s' % dn)
                cos_template.delete()
            except ldap.NO_SUCH_OBJECT:
                # Already deleted
                pass
            try:
                # Delete definition
                cos_pointer_def = CosPointerDefinition(self._instance, 'cn=nsPwPolicy_CoS,%s' % dn)
                cos_pointer_def.delete()
            except ldap.NO_SUCH_OBJECT:
                # Already deleted
                pass
        else:
            try:
                # Cleanup user entry
                entry.remove("pwdpolicysubentry", pwp_entry.dn)
            except ldap.NO_SUCH_ATTRIBUTE:
                # Policy already removed from user
                pass

        # Remove the local policy
        try:
            pwp_entry.delete()
        except ldap.NO_SUCH_OBJECT:
            # Already deleted
            pass

        try:
            pwp_container.delete()
        except (ldap.NOT_ALLOWED_ON_NONLEAF, ldap.NO_SUCH_OBJECT):
            # There are other policies still using this container, no problem
            pass

    def set_global_policy(self, properties):
        """Configure global password policy

        :param properties: A dictionary with password policy attributes
        :type properties: dict
        """

        modlist = []
        for attr, value in properties.items():
            modlist.append((attr, value))

        if len(modlist) > 0:
            config = Config(self._instance)
            config.replace_many(*modlist)
        else:
            raise ValueError("There are no password policies to set")


class PwPolicyEntry(DSLdapObject):
    """A single instance of a task entry

    :param instance: An instance
    :type instance: lib389.DirSrv
    :param dn: Entry DN
    :type dn: str
    """

    def __init__(self, instance, dn):
        super(PwPolicyEntry, self).__init__(instance, dn)
        self._rdn_attribute = 'cn'
        self._must_attributes = ['cn']
        self._create_objectclasses = ['top', 'ldapsubentry', 'passwordpolicy']
        self._protected = False

    def is_user_policy(self):
        """Check if the policy entry is a user password policy"""

        pwp_manager = PwPolicyManager(self._instance)
        cn = self.get_attr_val_utf8_l('cn')
        entrydn = cn.replace('cn=nspwpolicyentry_user,', '')

        return pwp_manager.is_user_policy(entrydn)

    def is_subtree_policy(self):
        """Check if the policy entry is a user password policy"""

        pwp_manager = PwPolicyManager(self._instance)
        cn = self.get_attr_val_utf8_l('cn')
        entrydn = cn.replace('cn=nspwpolicyentry_subtree,', '')

        return pwp_manager.is_subtree_policy(entrydn)


class PwPolicyEntries(DSLdapObjects):
    """DSLdapObjects that represents all password policy entries in container.

    :param instance: An instance
    :type instance: lib389.DirSrv
    :param basedn: Suffix DN
    :type basedn: str
    :param rdn: The DN that will be combined wit basedn
    :type rdn: str
    """

    def __init__(self, instance, basedn):
        super(PwPolicyEntries, self).__init__(instance)
        self._objectclasses = [
            'top',
            'ldapsubentry',
            'passwordpolicy'
        ]
        self._filterattrs = ['cn']
        self._childobject = PwPolicyEntry
        self._basedn = basedn




# --- BEGIN COPYRIGHT BLOCK ---
# Copyright (C) 2015 Red Hat, Inc.
# All rights reserved.
#
# License: GPL (version 3 or any later version).
# See LICENSE for details.
# --- END COPYRIGHT BLOCK ---

import io
import logging
import ldif
import ldap
import binascii
from ldap.cidict import cidict
import sys

from lib389.exceptions import MissingEntryError
from lib389._constants import *
from lib389.properties import *
from lib389.utils import (ensure_str, ensure_bytes, ensure_list_bytes, display_log_data)

MAJOR, MINOR, _, _, _ = sys.version_info

log = logging.getLogger(__name__)


class FormatDict(cidict):
    def __getitem__(self, name):
        if name in self:
            return ' '.join(cidict.__getitem__(self, name))
        return None


class Entry(object):
    """This class represents an LDAP Entry object.

        An LDAP entry consists of a DN and a list of attributes.
        Each attribute consists of a name and a *list* of values.
        String values will be rendered badly!
            ex. {
                'uid': ['user01'],
                'cn': ['User'],
                'objectlass': [ 'person', 'inetorgperson' ]
             }

        In python-ldap, entries are returned as a list of 2-tuples.
        Instance variables:
          dn - string - the string DN of the entry
          data - cidict - case insensitive dict of the attributes and values
    """
    # the ldif class base64 encodes some attrs which I would rather see in raw
    # form - to encode specific attrs as base64, add them to the list below
    base64_attrs = ['nsstate']

    def __init__(self, entrydata):
        """entrydata is the raw data returned from the python-ldap
        result method, which is:
            * a search result entry     -> (dn, {dict...} )
            * or a reference            -> (None, reference)
            * or None.

        If creating a new empty entry, data is the string DN.
        """
        self.ref = None
        if entrydata:
            if isinstance(entrydata, tuple):
                if entrydata[0] is None:
                    self.ref = entrydata[1]  # continuation reference
                else:
                    self.dn = entrydata[0]
                    self.data = cidict(entrydata[1])
            elif isinstance(entrydata, str):
                if '=' not in entrydata:
                    raise ValueError('Entry dn must contain "="')
                self.dn = entrydata
                self.data = cidict()
        else:
            self.data = cidict()
            self.dn = None

    def __bool__(self):
        """
        This allows us to do tests like if entry: returns false if there
        is no data, true otherwise
        """
        return self.data is not None and len(self.data) > 0

    def __eq__(self, other):
        """
        Compare if two Entry objects are the same.

        This is good for fast comparison of the data we have,
        but it relies on a few things.

        Each Entry must have been searched and retrieved with the
        same filter and attribute requirements.
        If that isn't the case, this will immediately fail.
        """
        # We could make this "strict" by forcing this to pull back all
        # the attributes of the DN, and the nsUniqueID.
        # Guard from accidents
        if not isinstance(other, Entry):
            return False
        # Check that our DN is the same.
        if self.dn != other.dn:
            return False
        # Bail fast if the keys don't match
        if set(self.getAttrs()) != set(other.getAttrs()):
            return False
        # Check the values of each key
        for key in self.getAttrs():
            if set(self.getValues(key)) != set(other.getValues(key)):
                return False
        return True

    def __ne__(self, other):
        """
        Compare if two Entry objects are the same.

        See __eq__ for a description of limitations.
        """
        return not self.__eq__(other)

    def hasAttr(self, name):
        """
        Return True if this entry has an attribute named name, False otherwise
        """
        if self.data is None:
            # Perhaps this should be an exception?
            raise Exception('Invalid data state in Entry')
        # We can't actually enforce this because cidict doesn't inherit Mapping
        # if not isinstance(self.data, collections.Mapping):
        #     raise Exception('Invalid data type for Entry')
        return ensure_str(name) in self.data

    def __getitem__(self, name):
        # This should probably return getValues?
        return self.__getattr__(name)

    def __getattr__(self, name):
        """
        If name is the name of an LDAP attribute, return the first
        value for that attribute - equivalent to getValue - this allows
        the use of
            entry.cn
        instead of
            entry.getValue('cn')
        This also allows us to return None if an attribute is not found
        rather than throwing an exception
        """
        if name == 'dn' or name == 'data':
            return self.__dict__.get(name, None)
        return self.getValue(name)

    def getValuesSet(self, keys):
        """
        returns a set of values based on keys
        """
        result = {}
        for k in keys:
            result[k] = self.data.get(k, [])
        return result

    def getValues(self, name):
        """Get the list (array) of values for the attribute named name"""
        return self.data.get(name, [])

    def getValue(self, name):
        """Get the first value for the attribute named name"""
        return self.data.get(name, [None])[0]

    def hasValue(self, name, val=None):
        """True if the given attribute is present and has the given value

            TODO: list comparison preserves order: should I use a set?
        """
        if not self.hasAttr(name):
            return False
        if not val:
            return True
        if isinstance(val, list):
            return val == self.data.get(name)
        if isinstance(val, tuple):
            return list(val) == self.data.get(name)
        return ensure_bytes(val) in self.data.get(name)

    def hasValueCase(self, name, val):
        """
        True if the given attribute is present and has the given value -
        ase insensitive value match
        """
        if not self.hasAttr(name):
            return False
        return val.lower() in [x.lower() for x in self.data.get(name)]

    def setValue(self, name, *value):
        """
        Value passed in may be a single value, several values,
         or a single sequence.
        For example:
           ent.setValue('name', 'value')
           ent.setValue('name', 'value1', 'value2', ..., 'valueN')
           ent.setValue('name', ['value1', 'value2', ..., 'valueN'])
           ent.setValue('name', ('value1', 'value2', ..., 'valueN'))
        Since *value is a tuple, we may have to extract a list or tuple
        from that tuple as in the last two examples above
        """
        if isinstance(value[0], list) or isinstance(value[0], tuple):
            self.data[name] = value[0]
        else:
            self.data[name] = value

    def getAttrs(self):
        if not self.data:
            return []
        return list(self.data.keys())

    def iterAttrs(self, attrsOnly=False):
        if attrsOnly:
            return self.data.keys()
        else:
            return self.data.items()

    setValues = setValue

    def toTupleList(self):
        """
        Convert the attrs and values to a list of 2-tuples.  The first
        element of the tuple is the attribute name.  The second element
        is either a single value or a list of values.
        """
        # For python3, we have to make sure EVERYTHING is a byte string.
        # Else everything EXPLODES
        lt = None
        if MAJOR >= 3:
            # This converts the dict to a list of tuples,
            lt = []
            for k in self.data.keys():
                # l here is the
                vals = None
                if isinstance(self.data[k], list) or isinstance(self.data[k], tuple):
                    vals = ensure_list_bytes(self.data[k])
                else:
                    vals = ensure_list_bytes([self.data[k]])
                lt.append((k, vals))
            # lt is now complete.
        else:
            lt = list(self.data.items())
        return lt

    def getref(self):
        return self.ref

    def __str__(self):
        """
        Convert the Entry to its LDIF representation
        """
        return self.__repr__()

    def update(self, dct):
        """Update passthru to the data attribute."""
        log.debug("updating dn: {}".format(self.dn))
        for k, v in list(dct.items()):
            if isinstance(v, list) or isinstance(v, tuple):
                self.data[k] = v
            else:
                self.data[k] = [v]
        log.debug("updated dn: {} with {}".format(self.dn, display_log_data(dct)))

    def __repr__(self):
        """Convert the Entry to its LDIF representation"""
        sio = io.StringIO()
        """
        what's all this then?  the unparse method will currently only accept
        a list or a dict, not a class derived from them.  self.data is a
        cidict, so unparse barfs on it.  I've filed a bug against
        python-ldap, but in the meantime, we have to convert to a plain old
        dict for printing.
        I also don't want to see wrapping, so set the line width really high
        (1000)
        """
        newdata = {}
        for k, v in self.data.items():
            newdata[k] = ensure_list_bytes(v)

        ldif.LDIFWriter(
            sio, Entry.base64_attrs, 1000).unparse(self.dn, newdata)
        return sio.getvalue()

    def bin2b64(self):
        """
        Convert any binary values in the entry to base 64
        """
        for attr, vals in self.iterAttrs():
            attr_vals = []
            for val in vals:
                try:
                    val.decode('ascii')
                    attr_vals.append(val)
                except:
                    # We have a binary value we need to convert
                    b64 = binascii.b2a_base64(val)
                    if b64[-1] == '\n':
                        # Strip off the newline
                        b64 = b64[:-1]

                    attr_vals.append(b64)

            self.data[attr] = attr_vals

    def getJSONEntry(self):
        """
        Return a JSON dictionary representation of the entry:

            dn - Entry DN
            attrs - entry's attributes

            {
                "dn": "uid=user,dc=example,dc=com",
                "attrs": {
                    "objectClass": [
                        "top",
                        "person",
                        "organizationalPerson",
                        "inetorgperson",
                        "inetuser"
                    ],
                    "givenName": [
                        "user"
                    ],
                    "uid": [
                        "user"
                    ],
                    "sn": [
                        "user lastname"
                    ],
                    "cn": [
                        "user name"
                    ]
                }
            }
        """
        # Need to convert binary values to base64 for our JSON representations
        self.bin2b64()

        entry = {'dn': self.dn, 'attrs': dict(self.data)}
        return entry

    def create(self, type=ENTRY_TYPE_PERSON, entry_dn=None, properties=None):
        """
        Return - eventually creating a person entry with the given dn and
                 pwd. binddn can be a lib389.Entry
        """
        if not entry_dn:
            raise ValueError("entry_dn is mandatory")

        if not type:
            raise ValueError("type is mandatory")

        ent = Entry(entry_dn)
        ent.setValues(ENTRY_OBJECTCLASS, ENTRY_TYPE_TO_OBJECTCLASS[type])
        ent.setValues(ENTRY_USERPASSWORD, properties.get(ENTRY_USERPASSWORD,
                                                         ""))
        ent.setValues(ENTRY_SN, properties.get(ENTRY_SN,
                                               "bind dn pseudo user"))
        ent.setValues(ENTRY_CN, properties.get(ENTRY_CN,
                                               "bind dn pseudo user"))
        if type == ENTRY_TYPE_INETPERSON:
            ent.setValues(ENTRY_UID, properties.get(ENTRY_UID,
                                                    "bind dn pseudo user"))

        try:
            self.add_s(ent)
        except ldap.ALREADY_EXISTS:
            log.warning("Entry %s already exists" % entry_dn)

        try:
            entry = self._test_entry(entry_dn, ldap.SCOPE_BASE)
            return entry
        except MissingEntryError:
            log.exception("This entry should exist!")
            raise

    def getAcis(self):
        if not self.hasAttr('aci'):
            # There should be a better way to do this? Perhaps
            # self search for the aci attr?
            return []
        self.acis = [EntryAci(self, a, verbose=False) for a in self.getValues('aci')]

        return self.acis


class EntryAci(object):
    """Breaks down an aci attribute string from 389, into a dictionary
    of terms and values. These values can then be manipulated, and
    subsequently rebuilt into an aci string.

    :param entry: An entry
    :type entry: lib389._entry.Entry
    :param rawaci: Aci in a raw form
    :type rawaci: str
    :param verbose: False by default
    :type verbose: bool
    """

    # We seperate the keys into 3 groups, and one group that has overlap.
    # This is so we can not only split the aci, but rebuild it from the
    # dictionary at a later point in time.

    # These are top level aci comoponent keys
    _keys = ['targetscope',
             'targetattrfilters',
             'targattrfilters',
             'targetfilter',
             'targetattr',
             'target',
             'version 3.0;']
    # These are the keys which are seperated by ; in the version 3.0 stanza.
    _v3keys = ['allow',
               'acl',
               'deny']
    # These are the keys which are used on the inside of a v3 allow statement
    # We have them defined, but don't currently use them.
    _v3innerkeys = ['roledn',
                    'userattr',
                    'ip',
                    'dns',
                    'dayofweek',
                    'timeofday',
                    'authmethod',
                    'userdn',
                    'groupdn']
    # These keys values are prefixed with ldap:///, so we need to know to
    # re-prefix ldap:/// onto the value when we rebuild the aci
    _urlkeys = ['target',
                'userdn',
                'groupdn',
                'roledn']

    def __init__(self, entry, rawaci, verbose=False):
        """
        """
        self.verbose = verbose
        self.entry = entry
        self._rawaci = ensure_str(rawaci)
        self.acidata = self._parse_aci(self._rawaci)

    def __eq__(self, other):

        # If we are from different entries, we should fail
        # Can this fail if we have different filters to get the acis though?
        if self.entry.dn != other.entry.dn:
            return False
        # Should we be checking the generated Aci incase it has changed?
        return self.getRawAci() == other.getRawAci()

    def __ne__(self, other):
        return not self.__eq__(other)

    def _format_term(self, key, value_dict):
        rawaci = ''
        if value_dict.get('equal', True):
            rawaci += '="'
        else:
            rawaci += '!="'
        if key in self._urlkeys:
            values = ['ldap:///%s' % x for x in value_dict['values']]
        else:
            values = value_dict['values']
        for value in values[:-1]:
            rawaci += "%s || " % value
        rawaci += values[-1]
        rawaci += '"'
        if self.verbose:
            print("_format_term: %s" % rawaci)
        return rawaci

    def getRawAci(self):
        """This method will rebuild an aci from the contents of the acidata
        dict found on the object.

        :returns: An aci attribute string.
        """

        # Rebuild the aci from the .acidata.
        rawaci = ''
        # For each key in the outer segment
        # Add a (key = val);. Depending on key format val:
        for key in self._keys:
            for value_dict in self.acidata[key]:
                rawaci += '(%s %s)' % (key, self._format_term(key, value_dict))
        # Now create the v3.0 aci part
        rawaci += "(version 3.0; "
        # This could be neater ...
        rawaci += 'acl "%s";' % self.acidata['acl'][0]['values'][0]
        for key in ['allow', 'deny']:
            if len(self.acidata[key]) > 0:
                rawaci += '%s (' % key
                for value in self.acidata[key][0]['values'][:-1]:
                    rawaci += '%s, ' % value
                rawaci += '%s)' % self.acidata[key][0]['values'][-1]
                rawaci += ('(%s);' % self.acidata["%s_raw_bindrules" %
                                                  key][0]['values'][-1])
        rawaci += ")"
        return ensure_bytes(rawaci)

    def _normalize_term(self, term):
        """Normalize the term, remove duplicate spaces around sensitive keywords,
        and remove spaces around "version 3.0 ;"
        :retuns: normalized term string
        """

        # First reduce multiple consecutive spaces
        new_term = ' '.join(term.split())

        # Our parsing of version 3.0; is strict, need to take special care
        if new_term.lower().startswith('version 3.0'):
            parts = new_term.split(';', 1)
            new_term = parts[0].rstrip() + ';' + parts[1]
        return new_term

    def _find_terms(self, aci):
        if self.verbose:
            print("_find_terms aci: %s" % aci)
            print("_find_terms aci: %s" % type(aci))
        lbr_list = []
        rbr_list = []
        depth = 0
        for i, char in enumerate(aci):
            if char == '(' and depth == 0:
                lbr_list.append(i)
            if char == '(':
                depth += 1
            if char == ')' and depth == 1:
                rbr_list.append(i)
            if char == ')':
                depth -= 1
        # Now build a set of terms.
        terms = []
        if self.verbose:
            print("_find_terms lbr_list" % lbr_list)
            print("_find_terms rbr_list" % rbr_list)
        for lb, rb in zip(lbr_list, rbr_list):
            terms.append(self._normalize_term(aci[lb + 1:rb]))
        if self.verbose:
            print("_find_terms terms: %s" % terms)
        return terms

    def _parse_term(self, key, term):
        wdict = {'values': [], 'equal': True}
        # Nearly all terms are = seperated
        # We make a dict that holds "equal" and an array of values
        pre, val = term.split('=', 1)
        val = val.replace('"', '')
        if pre.strip() == '!':
            wdict['equal'] = False
        else:
            wdict['equal'] = True
        wdict['values'] = val.split('||')
        if key in self._urlkeys:
            # We could replace ldap:/// in some attrs?
            wdict['values'] = ([x.replace('ldap:///', '')
                               for x in wdict['values']])
        wdict['values'] = [x.strip() for x in wdict['values']]

        if self.verbose:
            print("_parse_term: %s" % wdict)
        return wdict

    def _parse_bind_rules(self, subterm):
        # First, determine if there are extraneous braces wrapping the term.
        subterm = subterm.strip()
        if subterm[0] == '(' and subterm[-1] == ')':
            subterm = subterm[1:-1]
        # terms = subterm.split('and')
        """
        We could parse everything into nice structures, and then work with
        them.  Or we can just leave the bind rule alone, as a string. Let
        the human do it.  It comes down to cost versus reward.
        """

        if self.verbose:
            print("_parse_bind_rules: %s" % subterm)

        return [subterm]

    def _parse_version_3_0(self, rawacipart, data):
        # We have to do this because it's not the same as other term formats.
        terms = []
        # bindrules = []
        interms = rawacipart.split(';')
        interms = [x.strip() for x in interms]
        for iwork in interms:
            for j in self._v3keys + self._v3innerkeys:
                if iwork.startswith(j) and j == 'acl':
                    t = iwork.split(' ', 1)[1]
                    t = t.replace('"', '')
                    data[j].append({'values': [t]})
                if iwork.startswith(j) and (j == 'allow' or j == 'deny'):
                    first = iwork.index('(') + 1
                    second = iwork.index(')', first)
                    # This could likely be neater ...
                    data[j].append(
                        {'values': [x.strip()
                         for x in iwork[first:second].split(',')]})
                    subterm = iwork[second + 1:]
                    data["%s_raw_bindrules" % j].append({
                        'values': self._parse_bind_rules(subterm)
                    })

        if self.verbose:
            print("_parse_version_3_0: %s" % terms)
        return terms

    def _parse_aci(self, rawaci):
        aci = rawaci
        # depth = 0
        data = {
            'rawaci': rawaci,
            'allow_raw_bindrules': [],
            'deny_raw_bindrules': [],
            }
        for k in self._keys + self._v3keys:
            data[k] = []
        # We need to get a list of all the depth 0 ( and )
        terms = self._find_terms(aci)

        while len(terms) > 0:
            work = terms.pop()
            for k in self._keys + self._v3keys + self._v3innerkeys:
                if work.startswith(k):
                    aci = work.replace(k, '', 1)
                    if k == 'version 3.0;':
                        """
                        We pop more inner terms out, but we don't need to
                        parse them "now" they get added to the queue
                        """
                        terms += self._parse_version_3_0(aci, data)
                        continue
                    data[k].append(self._parse_term(k, aci))
                    break
        if self.verbose:
            print("_parse_aci: %s" % data)
        return data



# --- BEGIN COPYRIGHT BLOCK ---
# Copyright (C) 2017 Red Hat, Inc.
# All rights reserved.
#
# License: GPL (version 3 or any later version).
# See LICENSE for details.
# --- END COPYRIGHT BLOCK ---

from lib389.idm.directorymanager import DirectoryManager
from lib389.cli_base import _get_arg, CustomHelpFormatter


def password_change(inst, basedn, log, args):
    # Due to an issue, we can't use extended op, so we have to
    # submit the password directly to the field.
    password = _get_arg(args.password, msg="Enter new directory manager password", hidden=True, confirm=True)
    dm = DirectoryManager(inst)
    dm.change_password(password)


def create_parsers(subparsers):
    directory_manager_parser = subparsers.add_parser('directory_manager', help="Manage the Directory Manager account", formatter_class=CustomHelpFormatter)

    subcommands = directory_manager_parser.add_subparsers(help='action')

    password_change_parser = subcommands.add_parser('password_change', help="Changes the password of the Directory Manager account", formatter_class=CustomHelpFormatter)
    password_change_parser.set_defaults(func=password_change)
    # This is to put in a dummy attr that args can work with. We do this
    # because the actual test case will over-ride it, but it prevents
    # a user putting the pw on the cli.
    password_change_parser.set_defaults(password=None)





# --- BEGIN COPYRIGHT BLOCK ---
# Copyright (C) 2015 Red Hat, Inc.
# All rights reserved.
#
# License: GPL (version 3 or any later version).
# See LICENSE for details.
# --- END COPYRIGHT BLOCK ---
#
import os
import pwd
import logging
import pytest
from lib389._constants import *
from lib389.properties import *
from lib389 import DirSrv, Entry

logging.getLogger(__name__).setLevel(logging.DEBUG)
log = logging.getLogger(__name__)

TEST_DN = "uid=test,%s" % DEFAULT_SUFFIX
INSTANCE_PORT = 54321
INSTANCE_SERVERID = 'standalone'
INSTANCE_BACKUP = os.environ.get('BACKUPDIR', DEFAULT_BACKUPDIR)


class TopologyInstance(object):
    def __init__(self, instance):
        self.instance = instance


@pytest.fixture(scope='module')
def topology(request):
    instance = DirSrv(verbose=False)

    args = {SER_HOST: LOCALHOST,
            SER_PORT: INSTANCE_PORT,
            SER_SERVERID_PROP: INSTANCE_SERVERID}
    instance.allocate(args)

    if instance.exists():
        instance.delete()

    def fin():
        if instance.exists():
            instance.delete()
    request.addfinalizer(fin)

    return TopologyInstance(instance)


def _add_user(topology):
    """Add a user to the instance"""

    topology.instance.add_s(Entry((TEST_DN, {'objectclass':
                                             "top person".split(),
                                             'objectclass':
                                             "organizationalPerson",
                                             'objectclass': "inetOrgPerson",
                                             'uid': 'test',
                                             'sn': 'test',
                                             'cn': 'test'})))
    log.info('User %s was added' % TEST_DN)


def test_allocate(topology):
    """Test instance.allocate() function
    for correct state and arguments
    """

    log.info('Check instance state is allocated')
    assert topology.instance.state == DIRSRV_STATE_ALLOCATED

    log.info('Try to add user. Failure is excpected')
    with pytest.raises(Exception):
        _add_user(topology)

    if os.getuid() == 0:
        userid = DEFAULT_USER
    else:
        userid = pwd.getpwuid(os.getuid())[0]

    log.info('Verify the settings')
    assert topology.instance.host == LOCALHOST
    assert topology.instance.port == INSTANCE_PORT
    assert topology.instance.sslport is None
    assert topology.instance.binddn == DN_DM
    assert topology.instance.bindpw == PW_DM
    assert topology.instance.creation_suffix == DEFAULT_SUFFIX
    assert topology.instance.userid == userid
    assert topology.instance.serverid == INSTANCE_SERVERID
    assert topology.instance.groupid == topology.instance.userid
    assert topology.instance.backupdir == INSTANCE_BACKUP

    log.info('Check that we can change the settings of an allocated DirSrv')
    args = {SER_SERVERID_PROP: INSTANCE_SERVERID,
            SER_HOST: LOCALHOST,
            SER_PORT: INSTANCE_PORT,
            SER_ROOT_DN: "uid=foo"}
    topology.instance.allocate(args)

    log.info('Check instance state is allocated')
    assert topology.instance.state == DIRSRV_STATE_ALLOCATED

    log.info('Verify the settings')
    assert topology.instance.host == LOCALHOST
    assert topology.instance.port == INSTANCE_PORT
    assert topology.instance.sslport is None
    assert topology.instance.binddn == "uid=foo"
    assert topology.instance.bindpw == PW_DM
    assert topology.instance.creation_suffix == DEFAULT_SUFFIX
    assert topology.instance.userid == userid
    assert topology.instance.serverid == INSTANCE_SERVERID
    assert topology.instance.groupid == topology.instance.userid
    assert topology.instance.backupdir == INSTANCE_BACKUP

    log.info('Restore back the valid parameters and check')
    args = {SER_SERVERID_PROP: INSTANCE_SERVERID,
            SER_HOST: LOCALHOST,
            SER_PORT: INSTANCE_PORT}
    topology.instance.allocate(args)

    log.info('Check instance state is allocated')
    assert topology.instance.state == DIRSRV_STATE_ALLOCATED

    log.info('Verify the settings')
    assert topology.instance.host == LOCALHOST
    assert topology.instance.port == INSTANCE_PORT
    assert topology.instance.sslport is None
    assert topology.instance.binddn == DN_DM
    assert topology.instance.bindpw == PW_DM
    assert topology.instance.creation_suffix == DEFAULT_SUFFIX
    assert topology.instance.userid == userid
    assert topology.instance.serverid == INSTANCE_SERVERID
    assert topology.instance.groupid == topology.instance.userid
    assert topology.instance.backupdir == INSTANCE_BACKUP


def test_create(topology):
    """Test instance creation
    Check status before and after
    Try to add a user
    """

    log.info('Check instance state is allocated')
    assert topology.instance.state == DIRSRV_STATE_ALLOCATED

    topology.instance.create()

    log.info('Check instance state is offline')
    assert topology.instance.state == DIRSRV_STATE_OFFLINE

    log.info('Try to add user. Failure is excpected')
    with pytest.raises(Exception):
        _add_user(topology)


def test_open(topology):
    """Test instance opening
    Check status before and after
    Try to add a user
    """

    log.info('Check instance state is offline')
    assert topology.instance.state == DIRSRV_STATE_OFFLINE

    topology.instance.open()

    log.info('Check instance state is online')
    assert topology.instance.state == DIRSRV_STATE_ONLINE

    log.info('Try to add user. Success is excpected')
    _add_user(topology)


def test_close(topology):
    """Test instance closing
    Check status before and after
    Try to add a user
    """

    log.info('Check instance state is online')
    assert topology.instance.state == DIRSRV_STATE_ONLINE

    topology.instance.close()

    log.info('Check instance state is offline')
    assert topology.instance.state == DIRSRV_STATE_OFFLINE

    log.info('Try to add user. Failure is excpected')
    with pytest.raises(Exception):
        _add_user(topology)


def test_delete(topology):
    """Test instance deletion
    Check status before and after
    Try to add a user
    """

    log.info('Check instance state is offline')
    assert topology.instance.state == DIRSRV_STATE_OFFLINE

    topology.instance.delete()

    log.info('Check instance state is allocated')
    assert topology.instance.state == DIRSRV_STATE_ALLOCATED

    log.info('Try to add user. Failure is excpected')
    with pytest.raises(Exception):
        _add_user(topology)


if __name__ == "__main__":
    CURRENT_FILE = os.path.realpath(__file__)
    pytest.main('-s -v %s' % CURRENT_FILE)
